import sys
import os
import pkg_resources
import argparse
import time
from . import tasks


def main():
    """Main routine of the chado-tools package"""

    # Parse the supplied command line arguments
    arguments = parse_arguments(sys.argv)
    command = sys.argv[1]
    sub_command = ""
    if command in wrapper_commands():
        sub_command = sys.argv[2]

    if command in init_commands():

        # Set database connection parameters
        tasks.init(command)
    else:

        # Check database access
        start_time = time.time()
        connection_string = tasks.create_connection_string(arguments.config, arguments.dbname)
        if tasks.check_access(connection_string, sub_command):

            # Run the command
            tasks.run_command_with_arguments(command, sub_command, arguments, connection_string)

        # Print run time
        if arguments.verbose:
            print("Runtime: {0:.2f} s".format(time.time()-start_time))


def init_commands() -> dict:
    """Lists the available sub-commands of the 'chado' command for database initiation"""
    return {
        "init": "set the default connection parameters",
        "reset": "reset the default connection parameters to factory settings"
    }


def general_commands() -> dict:
    """Lists the available general sub-commands of the 'chado' command with corresponding descriptions"""
    return {
        "connect": "connect to a CHADO database for an interactive session",
        "query": "query a CHADO database and export the result to a text file"
    }


def wrapper_commands() -> dict:
    """Lists the available 'wrapper' sub-commands of the 'chado' command with corresponding descriptions"""
    return {
        "extract": "run a pre-compiled query against the CHADO database",
        "insert": "insert a new entity of a specified type into the CHADO database",
        "delete": "delete an entity of a specified type from the CHADO database",
        "import": "import data from file into the CHADO database",
        "export": "export data from the CHADO database to file",
        "execute": "execute a function defined in a CHADO database",
        "admin": "perform administrative tasks, such as creating or dumping a CHADO database"
    }


def admin_commands() -> dict:
    """Lists the available sub-commands of the 'chado admin' command with corresponding descriptions"""
    return {
        "create": "create a new CHADO database",
        "drop": "drop a CHADO database",
        "dump": "dump a CHADO database into an archive file",
        "restore": "restore a CHADO database from an archive file",
        "setup": "set up a blank CHADO database according to a given schema",
        "grant": "grant privileges for a CHADO database to a user/role",
        "revoke": "revoke privileges for a CHADO database from a user/role"
    }


def extract_commands() -> dict:
    """Lists the available sub-commands of the 'chado extract' command with corresponding descriptions"""
    return {
        "organisms": "list all organisms in the CHADO database",
        "cvterms": "list all CV terms in the CHADO database",
        "genedb_products": "list all products of transcripts in the CHADO database",
        "stats": "obtain statistics to updates in a CHADO database"
    }


def insert_commands() -> dict:
    """Lists the available sub-commands of the 'chado insert' command with corresponding descriptions"""
    return {
        "organism": "insert an organism into the CHADO database"
    }


def delete_commands() -> dict:
    """Lists the available sub-commands of the 'chado delete' command with corresponding descriptions"""
    return {
        "organism": "delete an organism from the CHADO database"
    }


def import_commands() -> dict:
    """Lists the available sub-commands of the 'chado import' command with corresponding descriptions"""
    return {
        "essentials": "import basic terms into the CHADO database (for setup)",
        "ontology": "import an ontology into the CHADO database",
        "fasta": "import sequences from a FASTA file into the CHADO database",
        "gff": "import genomic data from a GFF3 file into the CHADO database",
        "gaf": "import gene annotation data from a GAF file into the CHADO database"
    }


def export_commands() -> dict:
    """Lists the available sub-commands of the 'chado export' command with corresponding descriptions"""
    return {
        "fasta": "export genome/protein sequences from the CHADO database to a FASTA file",
        "gff": "export genomic data from the CHADO database to a GFF3 file",
        "gaf": "export gene annotation data from the CHADO database to a GAF file"
    }


def execute_commands() -> dict:
    """Lists the available sub-commands of the 'chado execute' command with corresponding descriptions"""
    return {
        "audit_backup": "backs up the audit tables to a separate schema"
    }


def parse_arguments(input_arguments: list) -> argparse.Namespace:
    """Defines the formal arguments of the 'chado' command and parses the actual arguments accordingly"""

    # Create a parser and add global formal arguments
    program_name = os.path.basename(input_arguments[0])
    parser = argparse.ArgumentParser(description="Tools to access CHADO databases",
                                     epilog="For detailed usage information type '" + program_name + " <command> -h'",
                                     prog=program_name, allow_abbrev=False)
    parser.add_argument("-v", "--version", help="show the version of the software and exit",
                        action='version', version=str(pkg_resources.get_distribution("chado-tools").version))

    # Add subparsers for all sub-commands
    subparsers = parser.add_subparsers()

    for command, description in init_commands().items():
        # Create subparser
        subparsers.add_parser(command, description=description, help=description)

    for command, description in general_commands().items():
        # Create subparser and add general and specific formal arguments
        sub = subparsers.add_parser(command, description=description, help=description)
        add_general_arguments(sub)
        add_arguments_by_command(command, sub)

    for command, description in wrapper_commands().items():
        # Create subparser and add specific formal arguments
        sub = subparsers.add_parser(command, description=description, help=description)
        add_arguments_by_command(command, sub)

    # Parse the actual arguments
    return parser.parse_args(input_arguments[1:])


def add_general_arguments(parser: argparse.ArgumentParser):
    """Defines general formal arguments (available to all sub-commands)"""
    parser.add_argument("-V", "--verbose", action="store_true", help="verbose mode")
    parser.add_argument("-c", "--config", default="", help="YAML file containing connection details")
    parser.add_argument("dbname", help="name of the database")


def add_general_extract_arguments(parser: argparse.ArgumentParser):
    """Defines general formal arguments for all sub-commands that export data from a database"""
    parser.add_argument("-H", "--include_header", action="store_true",
                        help="include header in CSV output (default: False)")
    parser.add_argument("-d", "--delimiter", default="\t",
                        help="Character delimiting fields in CSV output (default: tab)")
    parser.add_argument("-o", "--output_file", default="", help="file into which data are exported (default: stdout)")
    parser.add_argument("-F", "--format", default="csv", choices=["csv", "json"],
                        help="format of the file (default: csv)")


def add_arguments_by_command(command: str, parser: argparse.ArgumentParser):
    """Defines formal arguments for a specified sub-command"""
    if command == "connect":
        pass
    elif command == "admin":
        add_admin_arguments(parser)
    elif command == "query":
        add_query_arguments(parser)
    elif command == "execute":
        add_execute_arguments(parser)
    elif command == "extract":
        add_extract_arguments(parser)
    elif command == "insert":
        add_insert_arguments(parser)
    elif command == "delete":
        add_delete_arguments(parser)
    elif command == "import":
        add_import_arguments(parser)
    elif command == "export":
        add_export_arguments(parser)
    else:
        print("Command '" + parser.prog + "' is not available.")


def add_admin_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado admin' sub-command"""
    parser.epilog = "For detailed usage information type '" + parser.prog + " <command> -h'"
    subparsers = parser.add_subparsers()
    for command, description in admin_commands().items():
        # Create subparser and add general and specific formal arguments
        sub = subparsers.add_parser(command, description=description, help=description)
        add_general_arguments(sub)
        add_admin_arguments_by_command(command, sub)


def add_admin_arguments_by_command(command: str, parser: argparse.ArgumentParser):
    """Defines formal arguments for a specified sub-command of 'chado admin'"""
    if command == "create":
        pass
    elif command == "drop":
        pass
    elif command == "dump":
        add_dump_arguments(parser)
    elif command == "restore":
        add_restore_arguments(parser)
    elif command == "setup":
        add_setup_arguments(parser)
    elif command == "grant":
        add_grant_arguments(parser)
    elif command == "revoke":
        add_revoke_arguments(parser)
    else:
        print("Command '" + parser.prog + "' is not available.")


def add_dump_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado admin dump' sub-command"""
    parser.add_argument("archive", help="archive file to be created")


def add_restore_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado admin restore' sub-command"""
    parser.add_argument("archive", help="archive file")


def add_setup_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado admin setup' sub-command"""
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-s", "--schema", choices=["gmod", "basic", "audit", "audit_backup"], default="gmod",
                       help="Database schema (default: GMOD schema 1.31)")
    group.add_argument("-f", "--schema_file", default="", help="File with database schema")


def add_grant_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado admin grant' sub-command"""
    parser.add_argument("-r", "--role", required=True, help="Name of the role/user")
    parser.add_argument("-s", "--schema", help="Database schema (default: all)")
    parser.add_argument("-w", "--write", action="store_true", help="Grant read-write access (default: read-only)")


def add_revoke_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado admin revoke' sub-command"""
    parser.add_argument("-r", "--role", required=True, help="Name of the role/user")
    parser.add_argument("-s", "--schema", help="Database schema (default: all)")


def add_query_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado query' sub-command"""
    add_general_extract_arguments(parser)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-f", "--input_file", default="", help="file containing an SQL query")
    group.add_argument("-q", "--query", default="", help="SQL query")


def add_execute_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado execute' sub-command"""
    parser.epilog = "For detailed usage information type '" + parser.prog + " <command> -h'"
    subparsers = parser.add_subparsers()
    for command, description in execute_commands().items():
        # Create subparser and add general and specific formal arguments
        sub = subparsers.add_parser(command, description=description, help=description)
        add_general_arguments(sub)
        add_execute_arguments_by_command(command, sub)


def add_execute_arguments_by_command(command: str, parser: argparse.ArgumentParser):
    """Defines formal arguments for a specified sub-command of 'chado execute'"""
    if command == "audit_backup":
        add_execute_backup_arguments(parser)
    else:
        print("Command '" + parser.prog + "' is not available.")


def add_execute_backup_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado execute audit_backup' sub-command"""
    parser.add_argument("--date", required=True, help="date for maximum age of logs to remain in main audit tables, "
                                                      "format 'YYYYMMDD'")


def add_extract_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado extract' sub-command"""
    parser.epilog = "For detailed usage information type '" + parser.prog + " <command> -h'"
    subparsers = parser.add_subparsers()
    for command, description in extract_commands().items():
        # Create subparser and add general and specific formal arguments
        sub = subparsers.add_parser(command, description=description, help=description)
        add_general_arguments(sub)
        add_general_extract_arguments(sub)
        add_extract_arguments_by_command(command, sub)


def add_extract_arguments_by_command(command: str, parser: argparse.ArgumentParser):
    """Defines formal arguments for a specified sub-command of 'chado extract'"""
    if command == "organisms":
        add_extract_organisms_arguments(parser)
    elif command == "cvterms":
        add_extract_cvterms_arguments(parser)
    elif command == "genedb_products":
        add_extract_genedb_product_arguments(parser)
    elif command == "stats":
        add_extract_stats_arguments(parser)
    else:
        print("Command '" + parser.prog + "' is not available.")


def add_extract_organisms_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado extract organisms' sub-command"""
    parser.add_argument("--public_only", action="store_true", help="only extract public genomes (default: extract all)")
    parser.add_argument("--extract_version", action="store_true", help="extract the genome version, if available")


def add_extract_cvterms_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado extract cvterms' sub-command"""
    parser.add_argument("--vocabulary", help="restrict to a vocabulary, e.g. 'relationship'")
    parser.add_argument("--database", help="restrict to a database, e.g. 'GO'")


def add_extract_genedb_product_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado extract genedb_products' sub-command"""
    parser.add_argument("-a", "--abbreviation", dest="organism",
                        help="restrict to a certain organism, defined by its abbreviation/short name (default: all)")


def add_extract_stats_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado stats' sub-command"""
    parser.add_argument("-a", "--abbreviation", dest="organism",
                        help="restrict to a certain organism, defined by its abbreviation/short name (default: all)")
    parser.add_argument("--start_date", required=True, help="date for maximum age of updates, format 'YYYYMMDD'")
    parser.add_argument("--end_date", default="", help="date for minimum age of updates, format 'YYYYMMDD' "
                                                       "(default: today)")


def add_insert_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado insert' sub-command"""
    parser.epilog = "For detailed usage information type '" + parser.prog + " <command> -h'"
    subparsers = parser.add_subparsers()
    for command, description in insert_commands().items():
        # Create subparser and add general and specific formal arguments
        sub = subparsers.add_parser(command, description=description, help=description)
        add_general_arguments(sub)
        add_insert_arguments_by_command(command, sub)


def add_insert_arguments_by_command(command: str, parser: argparse.ArgumentParser):
    """Defines formal arguments for a specified sub-command of 'chado insert'"""
    if command == "organism":
        add_insert_organism_arguments(parser)
    else:
        print("Command '" + parser.prog + "' is not available.")


def add_insert_organism_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado insert organism' sub-command"""
    parser.add_argument("-g", "--genus", required=True, help="genus of the organism")
    parser.add_argument("-s", "--species", required=True, help="species of the organism")
    parser.add_argument("-i", "--infraspecific_name", help="infraspecific name (strain) of the organism")
    parser.add_argument("-a", "--abbreviation", required=True, help="abbreviation/short name of the organism")
    parser.add_argument("--common_name", help="common name of the organism (default: use abbreviation, if provided)")
    parser.add_argument("--comment", help="comment")
    parser.add_argument("--genome_version", type=int, help="version of the genome (integer)")


def add_delete_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado delete' sub-command"""
    parser.epilog = "For detailed usage information type '" + parser.prog + " <command> -h'"
    subparsers = parser.add_subparsers()
    for command, description in delete_commands().items():
        # Create subparser and add general and specific formal arguments
        sub = subparsers.add_parser(command, description=description, help=description)
        add_general_arguments(sub)
        add_delete_arguments_by_command(command, sub)


def add_delete_arguments_by_command(command: str, parser: argparse.ArgumentParser):
    """Defines formal arguments for a specified sub-command of 'chado delete'"""
    if command == "organism":
        add_delete_organism_arguments(parser)
    else:
        print("Command '" + parser.prog + "' is not available.")


def add_delete_organism_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado delete organism' sub-command"""
    parser.add_argument("-a", "--abbreviation", required=True, dest="organism",
                        help="abbreviation/short name of the organism")


def add_import_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado import' sub-command"""
    parser.epilog = "For detailed usage information type '" + parser.prog + " <command> -h'"
    subparsers = parser.add_subparsers()
    for command, description in import_commands().items():
        # Create subparser and add general and specific formal arguments
        sub = subparsers.add_parser(command, description=description, help=description)
        add_general_arguments(sub)
        add_import_arguments_by_command(command, sub)


def add_import_arguments_by_command(command: str, parser: argparse.ArgumentParser):
    """Defines formal arguments for a specified sub-command of 'chado import'"""
    if command == "essentials":
        pass
    elif command == "ontology":
        add_import_ontology_arguments(parser)
    elif command == "gff":
        add_import_gff_arguments(parser)
    elif command == "fasta":
        add_import_fasta_arguments(parser)
    elif command == "gaf":
        add_import_gaf_arguments(parser)
    else:
        print("Command '" + parser.prog + "' is not available.")


def add_import_ontology_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado import ontology' sub-command"""
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("-f", "--input_file", default="", help="file containing CV terms")
    group.add_argument("-u", "--input_url", default="", help="URL to a file containing CV terms")
    parser.add_argument("-A", "--database_authority", required=True,
                        help="database authority of the terms in the file, e.g. 'GO'")
    parser.add_argument("-F", "--format", default="obo", choices=["obo", "owl"],
                        help="format of the file (default: obo)")


def add_import_gff_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado import gff' sub-command"""
    parser.add_argument("-f", "--input_file", required=True, help="GFF3 input file")
    parser.add_argument("-a", "--abbreviation", required=True, dest="organism",
                        help="abbreviation/short name of the organism")
    parser.add_argument("--fasta", help="FASTA input file with sequences")
    parser.add_argument("-t", "--sequence_type", choices=["chromosome", "supercontig", "contig", "region"],
                        default="region", help="type of the FASTA sequences, if present (default: region)")
    parser.add_argument("--fresh_load", action="store_true",
                        help="load a genome from scratch (default: load an update to an existing genome)")
    parser.add_argument("--force", action="store_true",
                        help="in case of a fresh load, purge all existing features of the organism")
    parser.add_argument("--full_genome", action="store_true",
                        help="in case of an update, assume that the GFF file contains all features of the organism")


def add_import_fasta_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado import fasta' sub-command"""
    parser.add_argument("-f", "--input_file", required=True, help="FASTA input file")
    parser.add_argument("-a", "--abbreviation", required=True, dest="organism",
                        help="abbreviation/short name of the organism")
    parser.add_argument("-t", "--sequence_type", choices=["chromosome", "supercontig", "contig", "region"],
                        default="region", help="type of the sequences (default: region)")


def add_import_gaf_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado import gaf' sub-command"""
    parser.add_argument("-f", "--input_file", required=True, help="GFF3 input file")
    parser.add_argument("-a", "--abbreviation", required=True, dest="organism",
                        help="abbreviation/short name of the organism")
    parser.add_argument("-L", "--annotation_level", choices=["default", "gene", "transcript", "protein"],
                        default="default", help="level to which GO terms are related in the database (default: "
                                                "same level as in the input file)")


def add_export_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado export' sub-command"""
    parser.epilog = "For detailed usage information type '" + parser.prog + " <command> -h'"
    subparsers = parser.add_subparsers()
    for command, description in export_commands().items():
        # Create subparser and add general and specific formal arguments
        sub = subparsers.add_parser(command, description=description, help=description)
        add_general_arguments(sub)
        add_export_arguments_by_command(command, sub)


def add_export_arguments_by_command(command: str, parser: argparse.ArgumentParser):
    """Defines formal arguments for a specified sub-command of 'chado export'"""
    if command == "fasta":
        add_export_fasta_arguments(parser)
    elif command == "gff":
        add_export_gff_arguments(parser)
    elif command == "gaf":
        add_export_gaf_arguments(parser)
    else:
        print("Command '" + parser.prog + "' is not available.")


def add_export_fasta_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado export fasta' sub-command"""
    parser.add_argument("-f", "--output_file", required=True, help="FASTA output file")
    parser.add_argument("-a", "--abbreviation", required=True, dest="organism",
                        help="abbreviation/short name of the organism")
    parser.add_argument("-t", "--sequence_type", required=True, choices=["contigs", "genes", "proteins"],
                        help="type of the sequences to be exported")
    parser.add_argument("-r", "--release", help="name of the FASTA release")
    parser.add_argument("--extract_version", action="store_true", help="extract the genome version, if available")
    parser.add_argument("--include_obsolete", action="store_true", help="export all features, including obsoletes")


def add_export_gff_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado export gff' sub-command"""
    parser.add_argument("-f", "--output_file", required=True, help="GFF output file")
    parser.add_argument("-a", "--abbreviation", required=True, dest="organism",
                        help="abbreviation/short name of the organism")
    parser.add_argument("--export_fasta", action="store_true", help="export FASTA sequences along with annotations")
    parser.add_argument("--fasta_file", help="FASTA output file with sequences (default: paste to end of GFF file)")
    parser.add_argument("--include_obsolete", action="store_true", help="export all features, including obsoletes")


def add_export_gaf_arguments(parser: argparse.ArgumentParser):
    """Defines formal arguments for the 'chado export gff' sub-command"""
    parser.add_argument("-f", "--output_file", required=True, help="GAF output file")
    parser.add_argument("-a", "--abbreviation", required=True, dest="organism",
                        help="abbreviation/short name of the organism")
    parser.add_argument("-A", "--database_authority", required=True,
                        help="database from which the file is created, e.g. 'UniProtKB'")
    parser.add_argument("-L", "--annotation_level", choices=["default", "gene", "transcript", "protein"],
                        default="default", help="level to which GO terms are related in the output file (default: "
                                                "same level as in the database)")
    parser.add_argument("--include_obsolete", action="store_true", help="export all features, including obsoletes")



import copy
from typing import Union, List
import sqlalchemy.sql
import sqlalchemy.orm.attributes
import sqlalchemy.engine
import sqlalchemy.schema
import sqlalchemy.event
from . import utils
from .orm import base, general, cv, organism, pub, sequence, companalysis, audit


class DatabaseAccessClient(object):
    """Base class for access to a database"""

    def __init__(self, uri: str):
        """Constructor - connect to database"""
        self.uri = uri
        self.engine = sqlalchemy.create_engine(self.uri)                            # type: sqlalchemy.engine.Engine

    def __del__(self):
        """Destructor - disconnect from database"""
        # self.engine.dispose()
        pass


class DDLClient(DatabaseAccessClient):
    """Base class for all classes using DDL"""

    def __init__(self, uri: str):
        """Constructor - connect to database"""
        super().__init__(uri)
        self.connection = self.engine.connect()

    def __del__(self):
        """Destructor - disconnect from database"""
        self.connection.close()
        super().__del__()

    def schema_exists(self, schema: str) -> bool:
        """Checks if a given schema exists in a database"""
        schemata_table = sqlalchemy.text("information_schema.schemata")
        schemata_condition = sqlalchemy.text("schema_name=:schema")
        exists_query = sqlalchemy.exists().select_from(schemata_table).where(
            schemata_condition.bindparams(schema=schema)).select()
        return self.connection.execute(exists_query).scalar()

    def table_exists(self, table: str, schema: str) -> bool:
        """Checks if a given table exists in a given database schema"""
        tables_table = sqlalchemy.text("information_schema.tables")
        tables_condition = sqlalchemy.text("table_schema=:schema AND table_name=:table")
        exists_query = sqlalchemy.exists().select_from(tables_table).where(
            tables_condition.bindparams(schema=schema, table=table)).select()
        return self.connection.execute(exists_query).scalar()

    def table_inherits(self, child_table: str, parent_table: str) -> bool:
        """Checks if a given table inherits from a given parent table"""
        inheritance_table = sqlalchemy.text("pg_catalog.pg_inherits")
        inheritance_condition = sqlalchemy.text(r"inhparent=:parent\:\:regclass AND inhrelid=:child\:\:regclass")
        inherits_query = sqlalchemy.exists().select_from(inheritance_table).where(
            inheritance_condition.bindparams(parent=parent_table, child=child_table)).select()
        return self.connection.execute(inherits_query).scalar()

    def trigger_exists(self, schema: str, trigger_name: str) -> bool:
        """Checks if a trigger with a given name exists in a database"""
        triggers_table = sqlalchemy.text("information_schema.triggers")
        triggers_condition = sqlalchemy.text("trigger_schema=:schema AND trigger_name=:trigger_name")
        exists_query = sqlalchemy.exists().select_from(triggers_table).where(
            triggers_condition.bindparams(schema=schema, trigger_name=trigger_name)).select()
        return self.connection.execute(exists_query).scalar()

    def function_exists(self, schema: str, function_name: str) -> bool:
        """Checks if a given function exists in a database"""
        functions_table = sqlalchemy.text("information_schema.routines")
        functions_condition = sqlalchemy.text("routine_schema=:schema AND routine_name=:function")
        exists_query = sqlalchemy.exists().select_from(functions_table).where(
            functions_condition.bindparams(schema=schema, function=function_name)).select()
        return self.connection.execute(exists_query).scalar()

    def role_exists(self, role_name: str) -> bool:
        """Checks if a given role/user exists in a database"""
        roles_table = sqlalchemy.text("pg_catalog.pg_roles")
        roles_condition = sqlalchemy.text("rolname=:role_name")
        exists_query = sqlalchemy.exists().select_from(roles_table).where(
            roles_condition.bindparams(role_name=role_name)).select()
        return self.connection.execute(exists_query).scalar()

    def execute_ddl(self, statements: Union[str, List[str]]) -> None:
        """Executes DDL statement[s]"""
        if isinstance(statements, list):
            for statement in statements:
                ddl = sqlalchemy.schema.DDL(statement)
                ddl.execute(self.engine)
                print(statement)
        else:
            ddl = sqlalchemy.schema.DDL(statements)
            ddl.execute(self.engine)
            print(statements)

    def create(self):
        """Basis function defined for completeness of the API"""
        pass


class RolesClient(DDLClient):
    """Class for handling users/roles in a CHADO database"""

    def grant_or_revoke(self, rolename: str, specific_schema: str, read_write: bool, grant_access: bool):
        """Checks for role and schema existence and grants privileges"""
        # Check if the user exists
        if not self.role_exists(rolename):
            print("Role '" + rolename + "' does not exist. Create it before granting privileges.")
            return

        # Obtain a list of schemata
        if specific_schema:
            schemata = [specific_schema]
        else:
            schemata = ["public", "audit", "audit_backup", "graph"]

        # Loop over all schemata
        for schema in schemata:

            # Check if the schema exists
            if not self.schema_exists(schema) and schema != "public":
                print("Schema '" + schema + "' does not exist.")
                continue

            # Revoke/grant privileges
            commands = self.revoke_privileges_commands(rolename, schema)
            if grant_access:
                commands.extend(self.grant_privileges_commands(rolename, schema, read_write))
            self.execute_ddl(commands)

            # Print feedback
            if grant_access:
                print("Privileges on schema '" + schema + "' granted to role '" + rolename + "'")
            else:
                print("Privileges on schema '" + schema + "' revoked from role '" + rolename + "'")

    @staticmethod
    def grant_privileges_commands(rolename: str, schema: str, read_write: bool) -> list:
        """Creates commands for granting privileges for accessing database objects to a role/user"""
        x_role = "\"" + rolename + "\""
        x_schema = "\"" + schema + "\""
        statements = list()
        statements.append(" ".join(["GRANT USAGE ON SCHEMA", x_schema, "TO", x_role]))
        statements.append(" ".join(["GRANT SELECT ON ALL SEQUENCES IN SCHEMA", x_schema, "TO", x_role]))
        statements.append(" ".join(["GRANT SELECT ON ALL TABLES IN SCHEMA", x_schema, "TO", x_role]))
        if read_write:
            statements.append(" ".join(["GRANT EXECUTE ON ALL FUNCTIONS IN SCHEMA", x_schema, "TO", x_role]))
            statements.append(" ".join(["GRANT USAGE ON ALL SEQUENCES IN SCHEMA", x_schema, "TO", x_role]))
            statements.append(" ".join(["GRANT INSERT ON ALL TABLES IN SCHEMA", x_schema, "TO", x_role]))
            if not schema.startswith("audit"):
                statements.append(" ".join(["GRANT UPDATE ON ALL TABLES IN SCHEMA", x_schema, "TO", x_role]))
                statements.append(" ".join(["GRANT DELETE ON ALL TABLES IN SCHEMA", x_schema, "TO", x_role]))
        return statements

    @staticmethod
    def revoke_privileges_commands(rolename: str, schema: str) -> list:
        # Create the 'revoke' commands
        x_role = "\"" + rolename + "\""
        x_schema = "\"" + schema + "\""
        statements = list()
        statements.append(" ".join(["REVOKE USAGE ON SCHEMA", x_schema, "FROM", x_role]))
        statements.append(" ".join(["REVOKE ALL PRIVILEGES ON ALL FUNCTIONS IN SCHEMA", x_schema, "FROM", x_role]))
        statements.append(" ".join(["REVOKE ALL PRIVILEGES ON ALL SEQUENCES IN SCHEMA", x_schema, "FROM", x_role]))
        statements.append(" ".join(["REVOKE ALL PRIVILEGES ON ALL TABLES IN SCHEMA", x_schema, "FROM", x_role]))
        return statements


class SchemaSetupClient(DDLClient):
    """Base Class for setting up a CHADO database schema"""

    def __init__(self, uri: str):
        super().__init__(uri)
        self.base = base.PublicBase
        self.metadata = self.base.metadata
        self.schema = self.metadata.schema

    def create(self):
        """Main function for filling the schema with life"""

        # Create the tables
        self.metadata.create_all(self.engine, tables=self.metadata.sorted_tables)
        print("Created missing tables in schema '" + self.schema + "'.")

    def create_schema(self) -> None:
        """Creates a schema in the target database, if it doesn't exist yet"""
        if not self.schema_exists(self.schema):
            command = " ".join(["CREATE SCHEMA", self.schema])
            self.execute_ddl(command)
            print("Created schema '" + self.schema + "'.")
        else:
            print("Schema '" + self.schema + "' already exists.")

    def setup_inheritance(self, parent_table: str, child_tables: list) -> None:
        """Sets ups inheritance between parent table and child tables in a schema"""

        # Loop over all child tables
        full_parent_table = self.schema + '.' + parent_table
        for child_table in child_tables:
            full_child_table = self.schema + '.' + child_table

            # Check if the child table exists and if yes, if it already inherits from the parent table
            if self.table_exists(child_table, self.schema) and \
                    not self.table_inherits(full_child_table, full_parent_table):

                # Set up inheritance
                inherit_command = " ".join(["ALTER TABLE", full_child_table, "INHERIT", full_parent_table])
                self.execute_ddl(inherit_command)

    @staticmethod
    def create_trigger_function(function_schema: str, function_name: str, definition: str) -> str:
        return "CREATE OR REPLACE FUNCTION " + function_schema + "." + function_name + "()\n" \
               + "RETURNS trigger\nLANGUAGE plpgsql\nAS $function$\nBEGIN\n" \
               + definition + ";" \
               + "\nEND;\n$function$"

    @staticmethod
    def create_generic_trigger(trigger_name: str, function_schema: str, function_name: str, table_schema: str,
                               table_name: str) -> str:
        return "CREATE TRIGGER " + trigger_name \
               + " AFTER INSERT OR UPDATE OR DELETE ON " + table_schema + "." + table_name \
               + " FOR EACH ROW EXECUTE PROCEDURE " + function_schema + "." + function_name + "()"


class PublicSchemaSetupClient(SchemaSetupClient):
    """Class for setting up the general tables of a CHADO database"""

    def __init__(self, uri: str):
        """Constructor"""
        super().__init__(uri)
        self.modules = [general, cv, organism, pub, sequence, companalysis]


class AuditSchemaSetupClient(SchemaSetupClient):
    """Class for setting up an audit schema for a CHADO database"""

    def __init__(self, uri: str):
        """Constructor"""
        super().__init__(uri)
        self.modules = [audit]
        self.base = base.AuditBase
        self.metadata = self.base.metadata
        self.schema = self.metadata.schema
        self.master_table = audit.Audit.__table__

    def create(self):
        """Main function for filling the schema with life"""

        # Create the schema
        self.create_schema()

        # Create the tables
        public_client = PublicSchemaSetupClient(self.uri)
        data_tables = public_client.metadata.sorted_tables
        audit_tables = self.create_audit_tables(data_tables)
        audit_tablenames = [table.name for table in audit_tables]
        self.metadata.create_all(self.engine, tables=([self.master_table] + audit_tables))
        print("Created missing tables in schema '" + self.schema + "'.")

        # Set up inheritance
        self.setup_inheritance(self.master_table.name, audit_tablenames)
        print("Set up table inheritance in schema '" + self.schema + "'.")

        # Create triggers
        self.create_audit_triggers(audit_tables)
        print("Created audit triggers.")

    def create_audit_tables(self, data_tables: list) -> list:
        """Creates an audit table for each given data table"""

        # Define the audit base table and extract constraints
        audit_base_table = self.master_table                                            # type: sqlalchemy.Table
        audit_tables = []

        # Loop over all data tables
        for data_table in data_tables:                                                  # type: sqlalchemy.Table

            # Create a new audit table
            audit_table = sqlalchemy.Table(data_table.name, self.metadata)

            # Copy the columns of the audit base table to the new table
            for audit_column in audit_base_table.columns:                               # type: sqlalchemy.Column
                new_column = sqlalchemy.Column(audit_column.name, audit_column.type, nullable=audit_column.nullable,
                                               server_default=audit_column.server_default,
                                               primary_key=audit_column.primary_key)
                audit_table.append_column(new_column)

            # Copy the columns of the data table to the new table
            for data_column in data_table.columns:                                      # type: sqlalchemy.Column
                new_column = sqlalchemy.Column(data_column.name, data_column.type, nullable=data_column.nullable,
                                               server_default=data_column.server_default)
                audit_table.append_column(new_column)

            # Add table to list
            audit_tables.append(audit_table)

        # Return all audit tables
        return audit_tables

    def create_audit_triggers(self, audit_tables: list) -> None:
        """Creates the functions that trigger insertions of new lines in the audit table
        after each operation on the corresponding data table, and the triggers on the data tables"""

        # Retrieve columns of parent table as reference
        parent_column_names = [col.name for col in self.master_table.columns]

        # Loop over all audit tables
        for table in audit_tables:

            # Retrieve columns of this table; exclude columns inherited from parent table
            column_names = [col.name for col in table.columns]
            data_column_names = [name for name in column_names if name not in parent_column_names]

            # Define trigger function
            function_name = "public_" + table.name + "_proc"
            function_definition = self.generic_audit_function(self.schema, table.name, data_column_names)
            function_creator = self.create_trigger_function(self.schema, function_name, function_definition)

            # Create trigger function, if it doesn't exist yet
            if not self.function_exists(self.schema, function_name):
                self.execute_ddl(function_creator)

            # Define trigger
            trigger_name = table.name + "_audit_tr"
            trigger_creator = self.create_generic_trigger(trigger_name, self.schema, function_name,
                                                          "public", table.name)

            # Create trigger, if is doesn't exist yet
            if not self.trigger_exists("public", trigger_name):
                self.execute_ddl(trigger_creator)

    @staticmethod
    def generic_audit_function(schema: str, table: str, columns: list) -> str:
        return "IF TG_OP = 'INSERT' OR TG_OP = 'UPDATE' THEN\n" \
               + "\tINSERT INTO " + schema + "." + table + "(type, " + utils.list_to_string(columns, ", ") + ")" \
               + " VALUES (CAST(TG_OP AS " + base.operation_type.name + "), " \
               + utils.list_to_string(columns, ", ", "NEW") + ");\n" \
               + "\tRETURN NEW;\n" \
               + "ELSE\n" \
               + "\tINSERT INTO " + schema + "." + table + "(type, " + utils.list_to_string(columns, ", ") + ")" \
               + " VALUES (CAST(TG_OP AS " + base.operation_type.name + "), " \
               + utils.list_to_string(columns, ", ", "OLD") + ");\n" \
               + "\tRETURN OLD;\n" \
               + "END IF"


class AuditBackupSchemaSetupClient(AuditSchemaSetupClient):

    def __init__(self, uri: str):
        """Constructor"""
        super().__init__(uri)
        self.schema = "audit_backup"
        self.metadata = copy.deepcopy(self.base.metadata)
        self.metadata.schema = self.schema
        self.backup_master_table = self.master_table.tometadata(self.metadata, schema=self.schema)

    def create(self):
        """Main function for filling the schema with life"""

        # Create the schema
        self.create_schema()

        # Create the tables
        public_client = PublicSchemaSetupClient(self.uri)
        data_tables = public_client.metadata.sorted_tables
        audit_tables = self.create_audit_tables(data_tables)
        audit_tablenames = [table.name for table in audit_tables]
        self.metadata.create_all(self.engine, tables=([self.backup_master_table] + audit_tables))
        print("Created missing tables in schema '" + self.schema + "'.")

        # Set up inheritance
        self.setup_inheritance(self.backup_master_table.name, audit_tablenames)
        print("Set up table inheritance in schema '" + self.schema + "'.")

        # Create back-up function
        self.create_backup_function()
        print("Created a back_up function in schema '" + self.schema + "'.")

    def execute_backup_function(self, date: str):
        """Executes the function that backs up the audit tables"""
        if self.function_exists(self.schema, self.backup_function_name()):
            transaction = self.connection.begin()
            res = self.connection.execute(sqlalchemy.func.audit_backup.backup_proc(date)).scalar()
            transaction.commit()
            print("Moved " + str(res) + " database entries to schema " + self.schema)
        else:
            print("Function '" + self.schema + "." + self.backup_function_name() + "' does not exist")

    def create_backup_function(self):
        """Creates a function for moving audit tracks from the audit schema to the audit_backup schema"""
        if not self.function_exists(self.schema, self.backup_function_name()):
            function_name = self.schema + "." + self.backup_function_name()
            declarations = self.backup_function_declarations()
            definition = self.backup_function()
            backup_creator = self.backup_function_wrapper(function_name, declarations, definition)
            self.execute_ddl(backup_creator)

    @staticmethod
    def backup_function_name() -> str:
        return "backup_proc"

    @staticmethod
    def backup_function_wrapper(name: str, declarations: List[str], definition: str) -> str:
        return "CREATE OR REPLACE FUNCTION " + name + "(text)\n" \
               + "RETURNS bigint\nLANGUAGE plpgsql\nAS $function$\nDECLARE\n" \
               + ";\n".join(declarations) + ";\n" \
               + "BEGIN\n" \
               + definition + ";\n" \
               + "END;\n$function$"

    @staticmethod
    def backup_function_declarations() -> List[str]:
        return ["all_rows BIGINT", "table_rows BIGINT", "tabname RECORD"]

    @staticmethod
    def backup_function() -> str:
        return "\ttable_rows:=0;\n" \
               "\tall_rows:=0;\n" \
               "\tFOR tabname IN (SELECT table_name FROM information_schema.tables " \
               "WHERE table_schema = 'audit' AND table_name != 'audit') LOOP\n" \
               "\tEXECUTE\n" \
               "\t'INSERT INTO audit_backup.' || quote_ident(tabname.table_name) || " \
               "' (SELECT * FROM audit.' || quote_ident(tabname.table_name) || " \
               "' WHERE time < CAST($1 AS timestamp))'\n" \
               "\tUSING $1;\n" \
               "\tGET DIAGNOSTICS table_rows = ROW_COUNT;\n" \
               "\tall_rows:=all_rows+table_rows;\n" \
               "\tEXECUTE\n" \
               "\t'DELETE FROM audit.' || quote_ident(tabname.table_name) || ' WHERE time < CAST($1 AS timestamp)'\n" \
               "\tUSING $1;\n" \
               "\tEND LOOP;\n" \
               "\tRETURN all_rows"



from typing import List, Dict
import sqlalchemy.orm
from .. import utils, ddl
from ..orm import general, cv, pub, organism, sequence


class DatabaseError(Exception):
    pass


class InputFileError(Exception):
    pass


class IOClient(ddl.DatabaseAccessClient):
    """Base class for read-write access to a database"""

    def __init__(self, uri: str):
        """Constructor - connect to database"""
        super().__init__(uri)
        session_maker = sqlalchemy.orm.sessionmaker(bind=self.engine)
        self.session = session_maker()                                              # type: sqlalchemy.orm.Session

    def __del__(self):
        """Destructor - disconnect from database"""
        self.session.close()
        super().__del__()

    def query_table(self, table, **kwargs) -> sqlalchemy.orm.Query:
        """Creates a query on a database table from given keyword arguments"""
        query = self.session.query(table)
        if kwargs:
            query = query.filter_by(**kwargs)
        return query

    def query_all(self, table, **kwargs):
        """Helper class querying a table and returning all results"""
        return self.query_table(table, **kwargs).all()

    def query_first(self, table, **kwargs):
        """Helper class querying a table and returning the first result"""
        return self.query_table(table, **kwargs).first()

    def add_and_flush(self, obj):
        """Adds an entry to a database table"""
        self.session.add(obj)
        self.session.flush()

    def insert_into_table(self, table, **kwargs):
        """Creates an entry and inserts it into a database table"""
        obj = table(**kwargs)
        self.add_and_flush(obj)
        return obj

    def find_or_insert(self, table, **kwargs):
        """Returns one entry of a database table matching a query. If no matching entry exists, it is created."""
        entry = self.query_first(table, **kwargs)
        if not entry:
            entry = self.insert_into_table(table, **kwargs)
        return entry


class ChadoClient(IOClient):
    """Class for import/export operations on Chado databases"""

    def __init__(self, uri: str, verbose=False):
        """Constructor"""

        # Connect to database
        super().__init__(uri)

        # Set up printer
        self.printer = utils.VerbosePrinter(verbose)

    def query_feature_relationship_by_type(self, subject_id: int, type_ids: List[int]) -> sqlalchemy.orm.Query:
        """Creates a query to select entries with specific 'type_id' from the feature_relationship table"""
        return self.session.query(sequence.FeatureRelationship)\
            .join(cv.CvTerm, sequence.FeatureRelationship.type)\
            .filter(sequence.FeatureRelationship.subject_id == subject_id)\
            .filter(sequence.FeatureRelationship.type_id.in_(type_ids))

    def query_featureprop_by_type(self, feature_id: int, type_ids: List[int]) -> sqlalchemy.orm.Query:
        """Creates a query to select entries with specific 'type_id' from the featureprop table"""
        return self.session.query(sequence.FeatureProp)\
            .filter(sequence.FeatureProp.feature_id == feature_id)\
            .filter(sequence.FeatureProp.type_id.in_(type_ids))

    def query_feature_synonym_by_type(self, feature_id: int, type_ids: List[int]) -> sqlalchemy.orm.Query:
        """Creates a query to select entries related to a specific 'synonym.type_id' from the feature_synonym table"""
        return self.session.query(sequence.FeatureSynonym)\
            .join(sequence.Synonym, sequence.FeatureSynonym.synonym)\
            .filter(sequence.FeatureSynonym.feature_id == feature_id)\
            .filter(sequence.Synonym.type_id.in_(type_ids))

    def query_feature_cvterm_by_ontology(self, feature_id: int, ontology_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select entries related to a specific 'dbxref.db_id' from the feature_cvterm table"""
        return self.session.query(sequence.FeatureCvTerm)\
            .join(cv.CvTerm, sequence.FeatureCvTerm.cvterm)\
            .join(general.DbxRef, cv.CvTerm.dbxref)\
            .filter(sequence.FeatureCvTerm.feature_id == feature_id)\
            .filter(general.DbxRef.db_id == ontology_id)

    def query_feature_cvterm_by_ontology_and_organism(self, organism_id: int, ontology_id: int
                                                      ) -> sqlalchemy.orm.Query:
        """Creates a query to select ontology terms associated with feature of a given organism"""
        return self.session.query(sequence.FeatureCvTerm)\
            .join(sequence.Feature, sequence.FeatureCvTerm.feature)\
            .join(cv.CvTerm, sequence.FeatureCvTerm.cvterm)\
            .join(general.DbxRef, cv.CvTerm.dbxref)\
            .filter(sequence.Feature.organism_id == organism_id)\
            .filter(general.DbxRef.db_id == ontology_id)

    def query_parent_features(self, subject_id: int, type_ids: List[int]) -> sqlalchemy.orm.Query:
        """Creates a query to select the parent feature(s) of a given feature"""
        return self.session.query(sequence.Feature)\
            .select_from(sequence.FeatureRelationship)\
            .join(sequence.Feature, sequence.FeatureRelationship.object)\
            .filter(sequence.FeatureRelationship.type_id.in_(type_ids))\
            .filter(sequence.FeatureRelationship.subject_id == subject_id)\
            .order_by(sequence.Feature.uniquename)

    def query_child_features(self, object_id: int, type_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select the child features of a given feature"""
        return self.session.query(sequence.Feature)\
            .select_from(sequence.FeatureRelationship)\
            .join(sequence.Feature, sequence.FeatureRelationship.subject)\
            .filter(sequence.FeatureRelationship.type_id == type_id)\
            .filter(sequence.FeatureRelationship.object_id == object_id)\
            .order_by(sequence.Feature.uniquename)

    def query_features_by_srcfeature(self, sequence_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select the features located on a given sequence"""
        return self.session.query(sequence.Feature).select_from(sequence.FeatureLoc)\
            .join(sequence.Feature, sequence.FeatureLoc.feature)\
            .filter(sequence.FeatureLoc.srcfeature_id == sequence_id)\
            .order_by(sequence.FeatureLoc.fmin)

    def query_features_by_property_type(self, organism_id: int, type_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select features of a given organism that have certain properties"""
        return self.session.query(sequence.Feature)\
            .select_from(sequence.FeatureProp)\
            .join(sequence.Feature, sequence.FeatureProp.feature)\
            .filter(sequence.Feature.organism_id == organism_id)\
            .filter(sequence.FeatureProp.type_id == type_id)

    def query_features_by_type(self, organism_id: int, type_ids: List[int]) -> sqlalchemy.orm.Query:
        """Creates a query to select features of a given organism and type"""
        return self.session.query(sequence.Feature)\
            .filter(sequence.Feature.organism_id == organism_id)\
            .filter(sequence.Feature.type_id.in_(type_ids))

    def query_protein_features(self, organism_id: int, gene_type_id: int, part_of_id: int, derives_from_id: int
                               ) -> sqlalchemy.orm.Query:
        """Creates a query to select protein features of a given organism"""
        transcript_feature = sqlalchemy.orm.aliased(sequence.Feature, name="transcript_feature")
        protein_feature = sqlalchemy.orm.aliased(sequence.Feature, name="protein_feature")
        gene_feature = sqlalchemy.orm.aliased(sequence.Feature, name="gene_feature")
        transcript_gene_relationship = sqlalchemy.orm.aliased(
            sequence.FeatureRelationship, name="transcript_gene_relationship")
        protein_transcript_relationship = sqlalchemy.orm.aliased(
            sequence.FeatureRelationship, name="protein_transcript_relationship")
        return self.session.query(protein_feature)\
            .join(protein_transcript_relationship,
                  protein_transcript_relationship.subject_id == protein_feature.feature_id)\
            .join(transcript_feature, protein_transcript_relationship.object)\
            .join(transcript_gene_relationship,
                  transcript_gene_relationship.subject_id == transcript_feature.feature_id)\
            .join(gene_feature, transcript_gene_relationship.object)\
            .filter(protein_feature.organism_id == organism_id)\
            .filter(protein_transcript_relationship.type_id == derives_from_id)\
            .filter(transcript_gene_relationship.type_id == part_of_id)\
            .filter(gene_feature.type_id == gene_type_id)

    def query_feature_properties(self, feature_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select key-value pairs from the 'featureprop' table"""
        return self.session.query(cv.CvTerm.name, sequence.FeatureProp.value)\
            .select_from(sequence.FeatureProp)\
            .join(cv.CvTerm, sequence.FeatureProp.type)\
            .filter(sequence.FeatureProp.feature_id == feature_id)

    def query_feature_pubs(self, feature_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select entries from the 'pub' table associated with a given feature"""
        return self.session.query(pub.Pub.uniquename)\
            .select_from(sequence.FeaturePub)\
            .join(pub.Pub, sequence.FeaturePub.pub)\
            .filter(sequence.FeaturePub.feature_id == feature_id)

    def query_feature_dbxrefs(self, feature_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select dbxrefs associated with a given feature"""
        return self.session.query(general.Db.name, general.DbxRef.accession)\
            .select_from(sequence.FeatureDbxRef)\
            .join(general.DbxRef, sequence.FeatureDbxRef.dbxref)\
            .join(general.Db, general.DbxRef.db)\
            .filter(sequence.FeatureDbxRef.feature_id == feature_id)

    def query_feature_synonyms(self, feature_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select synonyms associated with a given feature"""
        return self.session.query(cv.CvTerm.name.label("type"), sequence.Synonym.name.label("synonym"))\
            .select_from(sequence.FeatureSynonym)\
            .join(sequence.Synonym, sequence.FeatureSynonym.synonym)\
            .join(cv.CvTerm, sequence.Synonym.type)\
            .filter(sequence.FeatureSynonym.feature_id == feature_id)

    def query_feature_ontology_terms(self, feature_id: int, ontology_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select ontology terms associated with a given feature"""
        return self.session.query(general.Db.name, general.DbxRef.accession)\
            .select_from(sequence.FeatureCvTerm)\
            .join(cv.CvTerm, sequence.FeatureCvTerm.cvterm)\
            .join(general.DbxRef, cv.CvTerm.dbxref)\
            .join(general.Db, general.DbxRef.db)\
            .filter(sequence.FeatureCvTerm.feature_id == feature_id)\
            .filter(general.Db.db_id == ontology_id)

    def query_feature_cvterm_properties(self, feature_cvterm_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select key-value pairs from the 'feature_cvtermprop' table"""
        return self.session.query(cv.CvTerm.name, sequence.FeatureCvTermProp.value)\
            .select_from(sequence.FeatureCvTermProp)\
            .join(cv.CvTerm, sequence.FeatureCvTermProp.type)\
            .filter(sequence.FeatureCvTermProp.feature_cvterm_id == feature_cvterm_id)

    def query_feature_cvterm_pubs(self, feature_cvterm_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select entries from the 'pub' table associated with a given feature_cvterm"""
        return self.session.query(pub.Pub.uniquename)\
            .select_from(sequence.FeatureCvTerm)\
            .join(pub.Pub, sequence.FeatureCvTerm.pub)\
            .filter(sequence.FeatureCvTerm.feature_cvterm_id == feature_cvterm_id)

    def query_feature_cvterm_secondary_pubs(self, feature_cvterm_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select entries from the 'pub' table associated with a given feature_cvterm"""
        return self.session.query(pub.Pub.uniquename)\
            .select_from(sequence.FeatureCvTermPub)\
            .join(pub.Pub, sequence.FeatureCvTermPub.pub)\
            .filter(sequence.FeatureCvTermPub.feature_cvterm_id == feature_cvterm_id)

    def query_feature_cvterm_dbxrefs(self, feature_cvterm_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select dbxrefs associated with a given feature_cvterm"""
        return self.session.query(general.Db.name, general.DbxRef.accession)\
            .select_from(sequence.FeatureCvTermDbxRef)\
            .join(general.DbxRef, sequence.FeatureCvTermDbxRef.dbxref)\
            .join(general.Db, general.DbxRef.db)\
            .filter(sequence.FeatureCvTermDbxRef.feature_cvterm_id == feature_cvterm_id)

    def query_feature_cvterm_ontology_terms(self, feature_cvterm_id: int, ontology_id: int
                                            ) -> sqlalchemy.orm.Query:
        """Creates a query to select ontology terms associated with a given feature_cvterm"""
        return self.session.query(general.Db.name, general.DbxRef.accession)\
            .select_from(sequence.FeatureCvTerm)\
            .join(cv.CvTerm, sequence.FeatureCvTerm.cvterm)\
            .join(general.DbxRef, cv.CvTerm.dbxref)\
            .join(general.Db, general.DbxRef.db)\
            .filter(sequence.FeatureCvTerm.feature_cvterm_id == feature_cvterm_id)\
            .filter(general.Db.db_id == ontology_id)

    def query_cvterm_namespace(self, cvterm_id: int) -> sqlalchemy.orm.Query:
        """Creates a query to select the namespace/vocabulary of a CV term"""
        return self.session.query(cv.Cv.name).select_from(cv.CvTerm)\
            .join(cv.Cv, cv.CvTerm.cv)\
            .filter(cv.CvTerm.cvterm_id == cvterm_id)

    def query_all_organisms(self, query_version: bool) -> sqlalchemy.orm.Query:
        """Creates a query to select organisms in the database"""
        if query_version:
            return self.session.query(organism.Organism.abbreviation, organism.Organism.genus,
                                      organism.Organism.species, organism.Organism.infraspecific_name.label("strain"),
                                      organism.Organism.common_name, organism.Organism.version)
        else:
            return self.session.query(organism.Organism.abbreviation, organism.Organism.genus,
                                      organism.Organism.species, organism.Organism.infraspecific_name.label("strain"),
                                      organism.Organism.common_name)

    def query_organisms_by_property_type(self, type_id: int, query_version: bool) -> sqlalchemy.orm.Query:
        """Creates a query to select organisms with a specific property in the database"""
        if query_version:
            query = self.session.query(organism.Organism.abbreviation, organism.Organism.genus,
                                       organism.Organism.species, organism.Organism.infraspecific_name.label("strain"),
                                       organism.Organism.common_name, organism.Organism.version)
        else:
            query = self.session.query(organism.Organism.abbreviation, organism.Organism.genus,
                                       organism.Organism.species, organism.Organism.infraspecific_name.label("strain"),
                                       organism.Organism.common_name)
        return query.select_from(organism.OrganismProp)\
            .join(organism.Organism, organism.OrganismProp.organism)\
            .filter(organism.OrganismProp.type_id == type_id)

    def _load_db(self, name: str) -> general.Db:
        """Loads a specific DB"""
        db_entry = self.query_first(general.Db, name=name)
        if not db_entry:
            raise DatabaseError("DB '" + name + "' not present in database")
        return db_entry

    def _load_dbs(self, names: List[str]) -> List[general.Db]:
        """Loads specific DBs"""
        db_entries = []
        for name in names:
            db_entries.append(self._load_db(name))
        return db_entries

    def _load_cvterm(self, term: str) -> cv.CvTerm:
        """Loads a specific CV term"""
        cvterm_entry = self.query_first(cv.CvTerm, name=term)
        if not cvterm_entry:
            raise DatabaseError("CV term '" + term + "' not present in database")
        return cvterm_entry

    def _load_cvterm_from_cv(self, term: str, vocabulary: str) -> cv.CvTerm:
        """Loads a specific CV term from a specific vocabulary"""
        cv_entry = self.query_first(cv.Cv, name=vocabulary)
        if not cv_entry:
            raise DatabaseError("CV '" + vocabulary + "' not present in database")
        cvterm_entry = self.query_first(cv.CvTerm, name=term, cv_id=cv_entry.cv_id)
        if not cvterm_entry:
            raise DatabaseError("CV term '" + term + "' not present in database")
        return cvterm_entry

    def _load_cvterms(self, terms: List[str]) -> List[cv.CvTerm]:
        """Loads specific CV terms"""
        cvterm_entries = []
        for term in terms:
            cvterm_entries.append(self._load_cvterm(term))
        return cvterm_entries

    def _load_terms_from_cv(self, vocabulary: str, relationship=False) -> List[cv.CvTerm]:
        """Loads CV terms from a given vocabulary and returns them in a list"""
        cv_entry = self.query_first(cv.Cv, name=vocabulary)
        if not cv_entry:
            raise DatabaseError("CV '" + vocabulary + "' not present in database")
        cvterm_entries = self.query_all(cv.CvTerm, cv_id=cv_entry.cv_id, is_relationshiptype=int(relationship))
        return cvterm_entries

    def _load_terms_from_cv_dict(self, vocabulary: str, required_terms: List[str], relationship=False
                                 ) -> Dict[str, cv.CvTerm]:
        """Loads CV terms from a given vocabulary and returns them in a dictionary, keyed by name"""
        cvterm_entries = self._load_terms_from_cv(vocabulary, relationship)
        cvterm_entries_dict = utils.list_to_dict(cvterm_entries, "name")
        for term in required_terms:
            if term not in cvterm_entries_dict:
                raise DatabaseError("CV term '" + term + "' not present in database")
        return cvterm_entries_dict

    @staticmethod
    def _extract_cvterm_ids_from_dict(cvterm_dict: Dict[str, cv.CvTerm], terms: List[str]) -> List[int]:
        """Extracts the IDs of CV terms from a given dictionary"""
        ids = []
        for term in terms:
            if term in cvterm_dict:
                ids.append(cvterm_dict[term].cvterm_id)
            else:
                raise DatabaseError("CV term '" + term + "' not present in database")
        return ids

    def _load_pub(self, pub_name: str) -> pub.Pub:
        """Loads a pub entry from the database"""
        pub_entry = self.query_first(pub.Pub, uniquename=pub_name)
        if not pub_entry:
            raise DatabaseError("Pub '" + pub_name + "' not present in database")
        return pub_entry

    def _load_organism(self, organism_name: str) -> organism.Organism:
        """Loads an organism entry from the database"""
        organism_entry = self.query_first(organism.Organism, abbreviation=organism_name)
        if not organism_entry:
            raise DatabaseError("Organism '" + organism_name + "' not present in database")
        return organism_entry

    def _load_feature_names(self, organism_entry: organism.Organism) -> List[str]:
        """Returns the IDs of all features for a given organism present in the database"""
        all_feature_names = []
        for feature_name, in self.session.query(sequence.Feature.uniquename).filter_by(
                organism_id=organism_entry.organism_id):
            all_feature_names.append(feature_name)
        return all_feature_names

    def _handle_db(self, new_entry: general.Db) -> general.Db:
        """Inserts or updates an entry in the 'db' table, and returns it"""

        # Check if the db is already present in the database
        existing_entry = self.query_first(general.Db, name=new_entry.name)
        if existing_entry:

            # Nothing to update, return existing entry
            return existing_entry
        else:

            # Insert new db entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted db '" + new_entry.name + "'")
            return new_entry

    def _handle_dbxref(self, new_entry: general.DbxRef, db_authority="") -> general.DbxRef:
        """Inserts or updates an entry in the 'dbxref' table, and returns it"""

        # Check if the dbxref is already present in the database
        existing_entry = self.query_first(general.DbxRef, db_id=new_entry.db_id, accession=new_entry.accession)
        if existing_entry:

            # Nothing to update, return existing entry
            return existing_entry
        else:

            # Insert new db entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted dbxref '" + db_authority + "." + new_entry.accession + "'")
            return new_entry

    def _handle_cv(self, new_entry: cv.Cv) -> cv.Cv:
        """Inserts or updates an entry in the 'cv' table, and returns it"""

        # Check if the cv is already present in the database
        existing_entry = self.query_first(cv.Cv, name=new_entry.name)
        if existing_entry:

            # Nothing to update, return existing entry
            return existing_entry
        else:

            # Insert new cv entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted controlled vocabulary '" + new_entry.name + "'")
            return new_entry

    def _handle_cvterm(self, new_entry: cv.CvTerm, vocabulary="") -> cv.CvTerm:
        """Inserts or updates an entry in the 'cvterm' table, and returns it"""

        # Check if the cvterm is already present in the database (look for dbxref)
        existing_entry = self.query_first(cv.CvTerm, dbxref_id=new_entry.dbxref_id)
        if existing_entry:

            # Nothing to update, return existing entry
            return existing_entry
        else:

            # Check if the cvterm is already present in the database (look for cv_id and name)
            existing_entry = self.query_first(cv.CvTerm, cv_id=new_entry.cv_id, name=new_entry.name)
            if existing_entry:

                # Nothing to update, return existing entry
                return existing_entry
            else:

                # Insert new cvterm entry
                self.add_and_flush(new_entry)
                self.printer.print("Inserted term '" + new_entry.name + "' in vocabulary '" + vocabulary + "'")
                return new_entry

    def _handle_feature(self, new_entry: sequence.Feature, organism_name="") -> sequence.Feature:
        """Inserts or updates an entry in the 'feature' table and returns it"""

        # Check if the feature is already present in the database
        existing_entry = self.query_first(sequence.Feature, organism_id=new_entry.organism_id,
                                          uniquename=new_entry.uniquename)
        if existing_entry:

            # Check if the entries in database and file have the same properties, and update if not
            if self.update_feature_properties(existing_entry, new_entry):
                self.printer.print("Updated feature '" + existing_entry.uniquename + "' for organism '"
                                   + organism_name + "'")
            return existing_entry
        else:

            # Insert new feature entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted feature '" + new_entry.uniquename + "' for organism '" + organism_name + "'")
            return new_entry

    def _handle_featureloc(self, new_entry: sequence.FeatureLoc, feature_name="") -> sequence.FeatureLoc:
        """Inserts or updates an entry in the 'featureloc' table, and returns it"""

        # Check if the featureloc is already present in the database
        existing_entry = self.query_first(sequence.FeatureLoc, feature_id=new_entry.feature_id)
        if existing_entry:

            # Check if the entries in database and file have the same properties, and update if not
            if self.update_featureloc_properties(existing_entry, new_entry):
                self.printer.print("Updated featureloc for feature '" + feature_name + "'")
            return existing_entry
        else:

            # Insert new featureloc entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted featureloc for feature '" + feature_name + "'")
            return new_entry

    def _handle_featureprop(self, new_entry: sequence.FeatureProp, existing_entries: List[sequence.FeatureProp],
                            property_name="", value="", feature_name="") -> sequence.FeatureProp:
        """Inserts or updates an entry in the 'featureprop' table, and returns it"""

        # Check if the featureprop is already present in the database
        matching_entries = utils.filter_objects(existing_entries, type_id=new_entry.type_id)
        for matching_entry in matching_entries:

            # Check if the entries in database and file have the same properties
            if matching_entry.value == new_entry.value:

                # Nothing to update; return existing entry
                return matching_entry
            else:

                # Adjust 'rank' to avoid a violation of the UNIQUE constraint
                new_entry.rank = max(new_entry.rank, matching_entry.rank+1)
        else:

            # Insert new featureprop entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted property '" + property_name + "' = '" + value + "' for feature '"
                               + feature_name + "'")
            return new_entry

    def _delete_featureprop(self, new_entries: List[sequence.FeatureProp],
                            existing_entries: List[sequence.FeatureProp], feature_name=""
                            ) -> List[sequence.FeatureProp]:
        """Deletes entries from the 'featureprop' table"""

        # Loop over all entries in the database
        deleted_entries = []
        for existing_entry in existing_entries:

            # Check if the entry is also present in the file
            matching_entries = utils.filter_objects(new_entries, type_id=existing_entry.type_id,
                                                    rank=existing_entry.rank)
            if not matching_entries:

                # Delete the entry from the database
                self.session.delete(existing_entry)
                deleted_entries.append(existing_entry)
                type_entry = self.query_first(cv.CvTerm, cvterm_id=existing_entry.type_id)
                self.printer.print("Deleted property '" + type_entry.name + "' = '" + existing_entry.value
                                   + "' for feature '" + feature_name + "'")

        return deleted_entries

    def _handle_feature_dbxref(self, new_entry: sequence.FeatureDbxRef, existing_entries: List[sequence.FeatureDbxRef],
                               crossref="", feature_name="") -> sequence.FeatureDbxRef:
        """Inserts or updates an entry in the 'feature_dbxref' table, and returns it"""

        # Check if the feature_dbxref is already present in the database
        matching_entries = utils.filter_objects(existing_entries, dbxref_id=new_entry.dbxref_id)
        if matching_entries:

            # Check if the entries in database and file have the same properties (there can only be one)
            matching_entry = matching_entries[0]
            if self.update_feature_dbxref_properties(matching_entry, new_entry):
                self.printer.print("Updated cross reference '" + crossref + "' for feature '" + feature_name + "'")
            return matching_entry
        else:

            # Insert new feature_dbxref entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted cross reference '" + crossref + "' for feature '" + feature_name + "'")
            return new_entry

    def _delete_feature_dbxref(self, new_entries: List[sequence.FeatureDbxRef],
                               existing_entries: List[sequence.FeatureDbxRef], feature_name=""
                               ) -> List[sequence.FeatureDbxRef]:
        """Deletes entries from the 'feature_dbxref' table"""

        # Loop over all entries in the database
        deleted_entries = []
        for existing_entry in existing_entries:

            # Check if the entry is also present in the file
            matching_entries = utils.filter_objects(new_entries, dbxref_id=existing_entry.dbxref_id)
            if not matching_entries:
                # Delete the entry from the database
                self.session.delete(existing_entry)
                deleted_entries.append(existing_entry)
                dbxref_entry = self.query_first(general.DbxRef, dbxref_id=existing_entry.dbxref_id)
                db_entry = self.query_first(general.Db, db_id=dbxref_entry.db_id)
                crossref = db_entry.name + ":" + dbxref_entry.accession
                self.printer.print("Deleted cross reference '" + crossref + "' for feature '" + feature_name + "'")

        return deleted_entries

    def _handle_feature_cvterm(self, new_entry: sequence.FeatureCvTerm, existing_entries: List[sequence.FeatureCvTerm],
                               term="", feature_name="") -> sequence.FeatureCvTerm:
        """Inserts or updates an entry in the 'feature_cvterm' table, and returns it"""

        # Check if the feature_cvterm is already present in the database
        matching_entries = utils.filter_objects(existing_entries, cvterm_id=new_entry.cvterm_id)
        if matching_entries:

            # Check if the entries in database and file have the same properties, and update if not
            # Note that there are potentially multiple entries (for different pub_ids/ranks). Ignored here.
            matching_entry = matching_entries[0]
            if self.update_feature_cvterm_properties(matching_entry, new_entry):
                self.printer.print("Updated CV term CV term '" + term + "' for feature '" + feature_name + "'")
            return matching_entry
        else:

            # Insert new feature_cvterm entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted CV term '" + term + "' for feature '" + feature_name + "'")
            return new_entry

    def _delete_feature_cvterm(self, new_entries: List[sequence.FeatureCvTerm],
                               existing_entries: List[sequence.FeatureCvTerm], feature_name=""
                               ) -> List[sequence.FeatureCvTerm]:
        """Deletes entries from the 'feature_cvterm' table"""

        # Loop over all entries in the database
        deleted_entries = []
        for existing_entry in existing_entries:

            # Check if the entry is also present in the file
            matching_entries = utils.filter_objects(new_entries, cvterm_id=existing_entry.cvterm_id,
                                                    pub_id=existing_entry.pub_id, rank=existing_entry.rank)
            if not matching_entries:

                # Delete the entry from the database
                self.session.delete(existing_entry)
                deleted_entries.append(existing_entry)
                cvterm_entry = self.query_first(cv.CvTerm, cvterm_id=existing_entry.cvterm_id)
                self.printer.print("Deleted CV term '" + cvterm_entry.name + "' for feature '" + feature_name + "'")

        return deleted_entries

    def _handle_feature_relationship(self, new_entry: sequence.FeatureRelationship,
                                     existing_entries: List[sequence.FeatureRelationship],
                                     subject_name="", object_name="", type_name="") -> sequence.FeatureRelationship:
        """Inserts or updates an entry in the 'feature_relationship' table, and returns it"""

        # Check if the feature_relationship is already present in the database
        matching_entries = utils.filter_objects(existing_entries, type_id=new_entry.type_id,
                                                object_id=new_entry.object_id)
        if matching_entries:

            # Check if the entries in database and file have the same properties, and update if not
            # Note that there are potentially multiple entries (for different ranks). Ignored here.
            matching_entry = matching_entries[0]
            if self.update_feature_relationship_properties(matching_entry, new_entry):
                self.printer.print("Updated relationship: '" + subject_name + "', '" + type_name + "', '"
                                   + object_name + "'")
            return matching_entry
        else:

            # Insert new feature_relationship entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted relationship: '" + subject_name + "', '" + type_name + "', '"
                               + object_name + "'")
            return new_entry

    def _delete_feature_relationship(self, new_entries: List[sequence.FeatureRelationship],
                                     existing_entries: List[sequence.FeatureRelationship], subject_name=""
                                     ) -> List[sequence.FeatureRelationship]:
        """Deletes entries from the 'feature_relationship' table"""

        # Loop over all entries in the database
        deleted_entries = []
        for existing_entry in existing_entries:

            # Check if the entry is also present in the file
            matching_entries = utils.filter_objects(new_entries, type_id=existing_entry.type_id,
                                                    object_id=existing_entry.object_id, rank=existing_entry.rank)
            if not matching_entries:

                # Delete the entry from the database
                self.session.delete(existing_entry)
                deleted_entries.append(existing_entry)
                object_entry = self.query_first(sequence.Feature, feature_id=existing_entry.object_id)
                type_entry = self.query_first(cv.CvTerm, cvterm_id=existing_entry.type_id)
                self.printer.print("Deleted relationship: '" + subject_name + "', '" + type_entry.name + "', '"
                                   + object_entry.uniquename + "'")

        return deleted_entries

    def _handle_synonym(self, new_entry: sequence.Synonym) -> sequence.Synonym:
        """Inserts or updates an entry in the 'synonym' table, and returns it"""

        # Check if the synonym is already present in the database
        existing_entry = self.query_first(sequence.Synonym, name=new_entry.name, type_id=new_entry.type_id)
        if existing_entry:

            # Check if the entries in database and file have the same properties, and update if not
            if self.update_synonym_properties(existing_entry, new_entry):
                self.printer.print("Updated synonym '" + existing_entry.name + "'.")
            return existing_entry
        else:

            # Insert a new synonym entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted synonym '" + new_entry.name + "'.")
            return new_entry

    def _handle_feature_synonym(self, new_entry: sequence.FeatureSynonym,
                                existing_entries: List[sequence.FeatureSynonym], synonym="", feature_name=""
                                ) -> sequence.FeatureSynonym:
        """Inserts or updates an entry in the 'feature_synonym' table, and returns it"""

        # Check if the feature_synonym is already present in the database
        matching_entries = utils.filter_objects(existing_entries, synonym_id=new_entry.synonym_id)
        if matching_entries:

            # Check if the entries in database and file have the same properties, and update if not
            # Note that there are potentially multiple entries (for different pub_ids). Ignored here.
            matching_entry = matching_entries[0]
            if self.update_feature_synonym_properties(matching_entry, new_entry):
                self.printer.print("Updated synonym '" + synonym + "' for feature '" + feature_name + "'")
            return matching_entry
        else:

            # Insert a new feature_synonym entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted synonym '" + synonym + "' for feature '" + feature_name + "'")
            return new_entry

    def _delete_feature_synonym(self, new_entries: List[sequence.FeatureSynonym],
                                existing_entries: List[sequence.FeatureSynonym], feature_name=""
                                ) -> List[sequence.FeatureSynonym]:
        """Deletes entries from the 'feature_synonym' table"""

        # Loop over all entries in the database
        deleted_entries = []
        for existing_entry in existing_entries:

            # Check if the entry is also present in the file
            matching_entries = utils.filter_objects(new_entries, synonym_id=existing_entry.synonym_id,
                                                    pub_id=existing_entry.pub_id)
            if not matching_entries:

                # Delete the entry from the database
                self.session.delete(existing_entry)
                deleted_entries.append(existing_entry)
                synonym_entry = self.query_first(sequence.Synonym, synonym_id=existing_entry.synonym_id)
                self.printer.print("Deleted synonym '" + synonym_entry.name + "' for feature '" + feature_name + "'")

        return deleted_entries

    def _handle_pub(self, new_entry: pub.Pub) -> pub.Pub:
        """Inserts or updates an entry in the 'pub' table, and returns it"""

        # Check if the publication is already present in the database
        existing_entry = self.query_first(pub.Pub, uniquename=new_entry.uniquename)
        if existing_entry:

            # Check if the entries in database and file have the same properties, and update if not
            if self.update_pub_properties(existing_entry, new_entry):
                self.printer.print("Updated publication '" + new_entry.uniquename + "'")
            return existing_entry
        else:

            # Insert a new feature_synonym entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted publication '" + new_entry.uniquename + "'")
            return new_entry

    def _handle_feature_pub(self, new_entry: sequence.FeaturePub, existing_entries: List[sequence.FeaturePub],
                            feature_name="", publication="") -> sequence.FeaturePub:
        """Inserts or updates an entry in the 'feature_pub' table, and returns it"""

        # Check if the feature_pub is already present in the database
        matching_entries = utils.filter_objects(existing_entries, pub_id=new_entry.pub_id)
        if matching_entries:

            # Nothing to update, return existing entry
            matching_entry = matching_entries[0]
            return matching_entry
        else:

            # Insert a new feature_pub entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted publication '" + publication + "' for feature '" + feature_name + "'")
            return new_entry

    def _delete_feature_pub(self, new_entries: List[sequence.FeaturePub],
                            existing_entries: List[sequence.FeaturePub], feature_name="") -> List[sequence.FeaturePub]:
        """Deletes entries from the 'feature_pub' table"""

        # Loop over all entries in the database
        deleted_entries = []
        for existing_entry in existing_entries:

            # Check if the entry is also present in the file
            matching_entries = utils.filter_objects(new_entries, pub_id=existing_entry.pub_id)
            if not matching_entries:

                # Delete the entry from the database
                self.session.delete(existing_entry)
                deleted_entries.append(existing_entry)
                pub_entry = self.query_first(pub.Pub,  pub_id=existing_entry.pub_id)
                self.printer.print("Deleted publication '" + pub_entry.uniquename
                                   + "' for feature '" + feature_name + "'")

        return deleted_entries

    def _handle_feature_cvtermprop(self, new_entry: sequence.FeatureCvTermProp,
                                   existing_entries: List[sequence.FeatureCvTermProp], key="", feature_name=""
                                   ) -> sequence.FeatureCvTermProp:
        """Inserts or updates an entry in the 'feature_cvtermprop' table, and returns it"""

        # Check if the feature_pub is already present in the database
        matching_entries = utils.filter_objects(existing_entries, type_id=new_entry.type_id)
        if matching_entries:

            # Check if the entries in database and file have the same properties, and update if not
            # Nothing to update, return existing entry
            matching_entry = matching_entries[0]
            if self.update_feature_cvtermprop_properties(matching_entry, new_entry):
                self.printer.print("Updated CV term property '" + key + "' = '" + new_entry.value
                                   + "' for feature '" + feature_name + "'")
            return matching_entry
        else:

            # Insert a new feature_cvtermprop entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted CV term property '" + key + "' = '" + new_entry.value
                               + "' for feature '" + feature_name + "'")
            return new_entry

    def _handle_feature_cvterm_dbxref(self, new_entry: sequence.FeatureCvTermDbxRef,
                                      existing_entries: List[sequence.FeatureCvTermDbxRef], crossref="", feature_name=""
                                      ) -> sequence.FeatureCvTermDbxRef:
        """Inserts or updates an entry in the 'feature_cvterm_dbxref' table, and returns it"""

        # Check if the feature_cvterm_dbxref is already present in the database
        matching_entries = utils.filter_objects(existing_entries, dbxref_id=new_entry.dbxref_id)
        if matching_entries:

            # Nothing to update, return existing entry
            matching_entry = matching_entries[0]
            return matching_entry
        else:

            # Insert a new feature_cvterm_dbxref entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted CV term cross reference '" + crossref + "' for feature '" + feature_name + "'")
            return new_entry

    def _handle_feature_cvterm_pub(self, new_entry: sequence.FeatureCvTermPub,
                                   existing_entries: List[sequence.FeatureCvTermPub], publication="", feature_name=""
                                   ) -> sequence.FeatureCvTermPub:
        """Inserts or updates an entry in the 'feature_cvterm_pub' table, and returns it"""

        # Check if the feature_cvterm_pub is already present in the database
        matching_entries = utils.filter_objects(existing_entries, pub_id=new_entry.pub_id)
        if matching_entries:

            # Nothing to update, return existing entry
            matching_entry = matching_entries[0]
            return matching_entry
        else:

            # Insert a new feature_cvterm_pub entry
            self.add_and_flush(new_entry)
            self.printer.print("Inserted CV term publication '" + publication + "' for feature '" + feature_name + "'")
            return new_entry

    def _mark_feature_as_obsolete(self, organism_entry: organism.Organism, uniquename: str) -> sequence.Feature:
        """Marks a feature as obsolete"""
        feature_entry = self.query_first(sequence.Feature, organism_id=organism_entry.organism_id,
                                         uniquename=uniquename)
        if not feature_entry.is_obsolete:
            feature_entry.is_obsolete = True
            self.printer.print("Marked feature '" + feature_entry.uniquename + "' as obsolete")
        return feature_entry

    @staticmethod
    def update_pub_properties(existing_entry: pub.Pub, new_entry: pub.Pub) -> bool:
        """Updates the properties of a pub entry in the database"""
        updated = False
        for attribute in ["title", "volume", "volumetitle", "series_name", "issue", "pyear", "pages", "miniref",
                          "publisher", "pubplace", "is_obsolete"]:
            if utils.copy_attribute(existing_entry, new_entry, attribute):
                updated = True
        return updated

    @staticmethod
    def update_feature_properties(existing_entry: sequence.Feature, new_entry: sequence.Feature) -> bool:
        """Updates the properties of a feature entry in the database"""
        updated = False
        for attribute in ["name", "type_id", "residues", "seqlen", "md5checksum", "is_analysis", "is_obsolete"]:
            if utils.copy_attribute(existing_entry, new_entry, attribute):
                updated = True
        return updated

    @staticmethod
    def update_featureloc_properties(existing_entry: sequence.FeatureLoc, new_entry: sequence.FeatureLoc) -> bool:
        """Updates the properties of a featureloc entry in the database"""
        updated = False
        for attribute in ["fmin", "fmax", "strand", "phase"]:
            if utils.copy_attribute(existing_entry, new_entry, attribute):
                updated = True
        return updated

    @staticmethod
    def update_feature_dbxref_properties(existing_entry: sequence.FeatureDbxRef,
                                         new_entry: sequence.FeatureDbxRef) -> bool:
        """Updates the properties of a feature_dbxref entry in the database"""
        updated = False
        if utils.copy_attribute(existing_entry, new_entry, "is_current"):
            updated = True
        return updated

    @staticmethod
    def update_feature_cvterm_properties(existing_entry: sequence.FeatureCvTerm,
                                         new_entry: sequence.FeatureCvTerm) -> bool:
        """Updates the properties of a feature_cvterm entry in the database"""
        updated = False
        if utils.copy_attribute(existing_entry, new_entry, "is_not"):
            updated = True
        return updated

    @staticmethod
    def update_feature_cvtermprop_properties(existing_entry: sequence.FeatureCvTermProp,
                                             new_entry: sequence.FeatureCvTermProp) -> bool:
        """Updates the properties of a feature_cvtermprop entry in the database"""
        updated = False
        if utils.copy_attribute(existing_entry, new_entry, "value"):
            updated = True
        return updated

    @staticmethod
    def update_feature_relationship_properties(existing_entry: sequence.FeatureRelationship,
                                               new_entry: sequence.FeatureRelationship) -> bool:
        """Updates the properties of a feature_relationship entry in the database"""
        updated = False
        if utils.copy_attribute(existing_entry, new_entry, "value"):
            updated = True
        return updated

    @staticmethod
    def update_synonym_properties(existing_entry: sequence.Synonym, new_entry: sequence.Synonym) -> bool:
        """Updates the properties of a synonym entry in the database"""
        updated = False
        if utils.copy_attribute(existing_entry, new_entry, "synonym_sgml"):
            updated = True
        return updated

    @staticmethod
    def update_feature_synonym_properties(existing_entry: sequence.FeatureSynonym,
                                          new_entry: sequence.FeatureSynonym) -> bool:
        """Updates the properties of a feature_synonym entry in the database"""
        updated = False
        for attribute in ["is_internal", "is_current"]:
            if utils.copy_attribute(existing_entry, new_entry, attribute):
                updated = True
        return updated



import sys
import os
import datetime
import subprocess
import urllib.request
import string
import random
import yaml


class EmptyObject:
    """Helper class that creates objects with attributes supplied by keyword arguments"""
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)


class VerbosePrinter:
    """Class printing messages if verbose argument is set"""

    def __init__(self, verbose: bool, separator=";"):
        """Constructor"""
        self.verbose = verbose
        self.separator = separator

    def print(self, message):
        """Prints a message if set to verbose. If the message is a list, the method prints all elements,
        separated by the set separator"""
        if self.verbose:
            if isinstance(message, list):
                print(*message, sep=self.separator)
            else:
                print(message)


def open_file_read(filename: str):
    """Function opening a (potentially gzipped) text file for read access"""
    if not filename:
        # Read from stdin
        f = sys.stdin
    else:
        # Check if file exists
        filepath = os.path.abspath(filename)
        if not os.path.exists(filepath):
            raise FileNotFoundError("File '" + filepath + "' does not exist.")
        # Open file for reading
        if filename.endswith(".gz"):
            subprocess.check_call("gunzip -t " + filename, shell=True, stderr=subprocess.DEVNULL)
            f = os.popen("gunzip -c " + filename, "r")
        else:
            f = open(filename, "r")
    return f


def open_file_write(filename: str):
    """Function opening a (potentially gzipped) text file for write access"""
    if not filename:
        # Write to stdout
        f = sys.stdout
    else:
        # Check if path exists
        filepath = os.path.abspath(os.path.dirname(filename))
        if not os.path.exists(filepath):
            raise FileNotFoundError("Directory '" + filepath + "' does not exist.")
        # Open file for writing
        if filename.endswith(".gz"):
            f = os.popen("gzip -9 -c > " + filename, "w")
        else:
            f = open(filename, "w")
    return f


def close(file):
    """Function closing a text file"""
    if file not in [sys.stdout, sys.stderr]:
        file.close()


def read_text(filename: str) -> str:
    """Function reading text from a file"""
    file = open_file_read(filename)
    content = file.read()
    close(file)
    return content


def write_text(filename: str, content: str) -> None:
    """Function writing text to a file"""
    file = open_file_write(filename)
    file.write(content)
    close(file)


def write_csv(filename: str, delimiter: str, content: list) -> None:
    """Function writing data arranged in a table into a CSV file"""
    file = open_file_write(filename)
    for row in content:
        file.write(list_to_string(row, delimiter) + "\n")
    close(file)


def parse_yaml(filename: str) -> dict:
    """Function parsing a YAML file"""
    stream = open_file_read(filename)
    data = yaml.load(stream)
    for key, value in data.items():
        if value is not None:
            data[key] = str(value).strip()
    close(stream)
    return data


def dump_yaml(filename: str, data: dict) -> None:
    """Function dumping data into a YAML file"""
    stream = open_file_write(filename)
    yaml.dump(data, stream)
    close(stream)


def list_to_string(the_list: list, delimiter: str, prefix=None) -> str:
    """Function concatenating all elements of a list"""
    the_string = []
    for element in the_list:
        if isinstance(element, bool) and element:
            the_string.append('t')
        elif isinstance(element, bool) and not element:
            the_string.append('f')
        elif isinstance(element, str):
            the_string.append(element)
        elif element is None:
            the_string.append("")
        else:
            the_string.append(str(element))
    if not prefix:
        return delimiter.join(the_string)
    else:
        return delimiter.join([prefix + "." + item for item in the_string])


def filter_objects(entries: list, **kwargs) -> list:
    """Filters a list of objects of any type according to given keyword arguments"""
    filtered_entries = []
    for entry in entries:
        for key, value in kwargs.items():
            if getattr(entry, key) != value:
                break
        else:
            filtered_entries.append(entry)
    return filtered_entries


def copy_attribute(old_object, new_object, attribute: str) -> bool:
    """Copies the value of a given attribute from one object to another"""
    new_value = getattr(new_object, attribute, None)
    old_value = getattr(old_object, attribute, None)
    if type(old_object) == type(new_object) and new_value is not None and old_value != new_value:
        setattr(old_object, attribute, new_value)
        return True
    return False


def list_to_dict(entries: list, key: str) -> dict:
    """Converts a list of objects of any type into a dictionary, using a specified object parameter as key"""
    dictionary = {}
    for entry in entries:
        current_key = getattr(entry, key)
        dictionary[current_key] = entry
    return dictionary


def random_string(n: int) -> str:
    """Generates a random string of n lowercase letters"""
    return "".join(random.choices(string.ascii_lowercase, k=n))


def random_integer(n: int) -> int:
    """Generates a random integer in the range [0, n]"""
    return random.randint(0, n)


def random_float() -> float:
    """Generates a random positive float number between 0 and 1"""
    n1 = random_integer(1000) + 1
    n2 = random_integer(1000) + 1
    if n2 > n1:
        return float(n1)/float(n2)
    else:
        return float(n2)/float(n1)


def current_date() -> str:
    """Function returning the current date in format 'YYYYMMDD"""
    return datetime.date.today().strftime('%Y%m%d')


def download_file(url: str) -> str:
    """Downloads a file from the internet"""
    print("Downloading file from URL " + url + " ...")
    file, headers = urllib.request.urlretrieve(url)
    return file



import unittest.mock
from .. import chado_tools, tasks, queries, dbutils, utils, ddl
from ..io import direct, essentials, ontology, fasta, gff, gaf


class TestTasks(unittest.TestCase):
    """Tests the functionality of all commands of the CHADO program"""

    def setUp(self):
        self.uri = tasks.create_connection_string("", "testdb")

    @unittest.mock.patch('pychado.utils.parse_yaml')
    @unittest.mock.patch('pychado.dbutils.default_configuration_file')
    def test_create_connection_string(self, mock_default, mock_parse):
        # Checks that the default configuration file is used unless another file is supplied
        self.assertIs(mock_default, dbutils.default_configuration_file)
        self.assertIs(mock_parse, utils.parse_yaml)

        mock_default.return_value = "default_file"
        tasks.create_connection_string("", "testdb")
        mock_default.assert_called()
        mock_parse.assert_called_with("default_file")

        mock_default.reset_mock()
        mock_parse.reset_mock()
        tasks.create_connection_string("user_supplied_file", "testdb")
        mock_default.assert_not_called()
        mock_parse.assert_called_with("user_supplied_file")

    @unittest.mock.patch('pychado.dbutils.exists')
    def test_access(self, mock_exist):
        # Checks that database access is only permitted if the database exists or gets created,
        # and checks that no existing database is overwritten
        self.assertIs(mock_exist, dbutils.exists)

        mock_exist.return_value = True
        self.assertTrue(tasks.check_access(self.uri, "connect"))
        self.assertFalse(tasks.check_access(self.uri, "create"))

        mock_exist.return_value = False
        self.assertFalse(tasks.check_access(self.uri, "connect"))
        self.assertTrue(tasks.check_access(self.uri, "create"))

    @unittest.mock.patch('pychado.dbutils.reset_default_parameters')
    @unittest.mock.patch('pychado.dbutils.set_default_parameters')
    def test_init_reset(self, mock_set, mock_reset):
        # Checks that the functions setting/resetting default connection parameters are correctly called
        self.assertIs(mock_set, dbutils.set_default_parameters)
        self.assertIs(mock_reset, dbutils.reset_default_parameters)

        tasks.init("init")
        mock_set.assert_called()
        mock_reset.assert_not_called()
        mock_set.reset_mock()
        mock_reset.reset_mock()

        tasks.init("reset")
        mock_set.assert_not_called()
        mock_reset.assert_called()
        mock_set.reset_mock()
        mock_reset.reset_mock()

        tasks.init("non_existing_command")
        mock_set.assert_not_called()
        mock_reset.assert_not_called()

    @unittest.mock.patch('pychado.dbutils.connect_to_database')
    def test_connect(self, mock_connect):
        # Checks that the function establishing a connection is correctly called
        self.assertIs(mock_connect, dbutils.connect_to_database)
        args = ["chado", "connect", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], "", parsed_args, self.uri)
        mock_connect.assert_called_with(self.uri)

    @unittest.mock.patch('pychado.dbutils.create_database')
    def test_create(self, mock_create):
        # Checks that the function creating and setting up a database is correctly called
        self.assertIs(mock_create, dbutils.create_database)
        args = ["chado", "admin", "create", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_create.assert_called_with(self.uri)

    @unittest.mock.patch('pychado.dbutils.drop_database')
    def test_drop(self, mock_drop):
        # Checks that a function dropping a database is correctly called
        self.assertIs(mock_drop, dbutils.drop_database)
        args = ["chado", "admin", "drop", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_drop.assert_called_with(self.uri)

    @unittest.mock.patch('pychado.dbutils.dump_database')
    def test_dump(self, mock_dump):
        # Checks that the function dumping a database is correctly called
        self.assertIs(mock_dump, dbutils.dump_database)
        args = ["chado", "admin", "dump", "testdb", "testarchive"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_dump.assert_called_with(self.uri, "testarchive")

    @unittest.mock.patch('pychado.dbutils.restore_database')
    @unittest.mock.patch('pychado.dbutils.create_database')
    def test_restore(self, mock_create, mock_restore):
        # Checks that the function restoring a database is correctly called
        self.assertIs(mock_create, dbutils.create_database)
        self.assertIs(mock_restore, dbutils.restore_database)
        args = ["chado", "admin", "restore", "testdb", "testarchive"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_create.assert_called_with(self.uri)
        mock_restore.assert_called_with(self.uri, "testarchive")

    @unittest.mock.patch('pychado.tasks.run_setup_command')
    def test_run_setup(self, mock_run):
        # Checks that database setup is correctly run
        self.assertIs(mock_run, tasks.run_setup_command)
        args = ["chado", "admin", "setup", "-f", "testschema", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_run.assert_called_with(parsed_args, self.uri)

    @unittest.mock.patch('pychado.ddl.AuditBackupSchemaSetupClient')
    @unittest.mock.patch('pychado.ddl.AuditSchemaSetupClient')
    @unittest.mock.patch('pychado.ddl.PublicSchemaSetupClient')
    @unittest.mock.patch('pychado.utils.download_file')
    @unittest.mock.patch('pychado.dbutils.setup_database')
    def test_setup(self, mock_setup, mock_download, mock_public_schema_client, mock_audit_schema_client,
                   mock_backup_schema_client):
        # Checks that the function setting up a database schema is correctly called
        self.assertIs(mock_setup, dbutils.setup_database)
        self.assertIs(mock_download, utils.download_file)
        self.assertIs(mock_public_schema_client, ddl.PublicSchemaSetupClient)
        self.assertIs(mock_audit_schema_client, ddl.AuditSchemaSetupClient)
        self.assertIs(mock_backup_schema_client, ddl.AuditBackupSchemaSetupClient)

        args = ["chado", "admin", "setup", "-f", "testschema", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_setup_command(parsed_args, self.uri)
        mock_download.assert_not_called()
        mock_setup.assert_called_with(self.uri, "testschema")

        mock_download.reset_mock()
        mock_download.return_value = "downloaded_schema"
        args = ["chado", "admin", "setup", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_setup_command(parsed_args, self.uri)
        mock_download.assert_called()
        mock_setup.assert_called_with(self.uri, "downloaded_schema")

        mock_public_schema_client.reset_mock()
        mock_audit_schema_client.reset_mock()
        mock_backup_schema_client.reset_mock()
        args = ["chado", "admin", "setup", "-s", "basic", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_setup_command(parsed_args, self.uri)
        mock_public_schema_client.assert_called_with(self.uri)
        mock_backup_schema_client.assert_not_called()
        mock_audit_schema_client.assert_not_called()
        self.assertIn(unittest.mock.call().create(), mock_public_schema_client.mock_calls)

        mock_public_schema_client.reset_mock()
        mock_audit_schema_client.reset_mock()
        mock_backup_schema_client.reset_mock()
        args = ["chado", "admin", "setup", "-s", "audit", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_setup_command(parsed_args, self.uri)
        mock_public_schema_client.assert_not_called()
        mock_audit_schema_client.assert_called_with(self.uri)
        mock_backup_schema_client.assert_not_called()
        self.assertIn(unittest.mock.call().create(), mock_audit_schema_client.mock_calls)

        mock_public_schema_client.reset_mock()
        mock_audit_schema_client.reset_mock()
        mock_backup_schema_client.reset_mock()
        args = ["chado", "admin", "setup", "-s", "audit_backup", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_setup_command(parsed_args, self.uri)
        mock_public_schema_client.assert_not_called()
        mock_audit_schema_client.assert_not_called()
        mock_backup_schema_client.assert_called_with(self.uri)
        self.assertIn(unittest.mock.call().create(), mock_backup_schema_client.mock_calls)

    @unittest.mock.patch('pychado.tasks.run_grant_revoke_command')
    def test_run_grant(self, mock_run):
        # Checks that database access granting is correctly run
        self.assertIs(mock_run, tasks.run_grant_revoke_command)
        args = ["chado", "admin", "grant", "-r", "testrole", "-s", "testschema", "-w", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_run.assert_called_with(parsed_args, self.uri, True)

        args = ["chado", "admin", "revoke", "-r", "testrole", "-s", "testschema", "testdb"]
        parsed_args = vars(chado_tools.parse_arguments(args))
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_run.assert_called_with(parsed_args, self.uri, False)

    @unittest.mock.patch('pychado.ddl.RolesClient')
    def test_grant(self, mock_client):
        # Checks that the function granting database access is correctly called
        self.assertIs(mock_client, ddl.RolesClient)
        args = ["chado", "admin", "grant", "-r", "testrole", "-s", "testschema", "-w", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_grant_revoke_command(parsed_args, self.uri, True)
        mock_client.assert_called_with(self.uri)
        self.assertIn(unittest.mock.call().grant_or_revoke("testrole", "testschema", True, True),
                      mock_client.mock_calls)

    @unittest.mock.patch('pychado.ddl.RolesClient')
    def test_revoke(self, mock_client):
        # Checks that the function revoking database access is correctly called
        self.assertIs(mock_client, ddl.RolesClient)
        args = ["chado", "admin", "revoke", "-r", "testrole", "-s", "testschema", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_grant_revoke_command(parsed_args, self.uri, False)
        mock_client.assert_called_with(self.uri)
        self.assertIn(unittest.mock.call().grant_or_revoke("testrole", "testschema", False, False),
                      mock_client.mock_calls)

    @unittest.mock.patch('pychado.tasks.run_query_command')
    def test_run_query(self, mock_run):
        # Checks that database queries are correctly run
        self.assertIs(mock_run, tasks.run_query_command)
        args = ["chado", "query", "-H", "-d", ";", "-o", "testfile", "-q", "testquery", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_run.assert_called_with(parsed_args, self.uri)

    @unittest.mock.patch('pychado.dbutils.query_and_print')
    @unittest.mock.patch('pychado.utils.read_text')
    def test_query(self, mock_read, mock_query):
        # Checks that the function querying a database is correctly called
        self.assertIs(mock_read, utils.read_text)
        self.assertIs(mock_query, dbutils.query_and_print)

        # Direct query
        args = ["chado", "query", "-H", "-d", ";", "-o", "testfile", "-F", "json", "-q", "testquery", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        mock_query.return_value = ['a', 'b', 'c']
        tasks.run_query_command(parsed_args, self.uri)
        mock_query.assert_called_with(self.uri, "testquery", "testfile", "json", True, ";")

        # Query extracted from file
        args = ["chado", "query", "-d", "\t", "-o", "otherfile", "-f", "testqueryfile", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        mock_read.return_value = "query_from_file"
        tasks.run_query_command(parsed_args, self.uri)
        mock_query.assert_called_with(self.uri, "query_from_file", "otherfile", "csv", False, "\t")

    @unittest.mock.patch('pychado.tasks.run_execute_command')
    def test_run_execute(self, mock_run):
        # Checks that calls to database functions are correctly run
        self.assertIs(mock_run, tasks.run_execute_command)
        args = ["chado", "execute", "audit_backup", "--date", "testdate", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_run.assert_called_with("audit_backup", parsed_args, self.uri)

    @unittest.mock.patch('pychado.ddl.AuditBackupSchemaSetupClient')
    def test_execute_audit_backup(self, mock_backup_schema_client):
        # Checks that the function calling the backup process of the audit schema is correctly called
        self.assertIs(mock_backup_schema_client, ddl.AuditBackupSchemaSetupClient)
        args = ["chado", "execute", "audit_backup", "--date", "testdate", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_execute_command(args[2], parsed_args, self.uri)
        mock_backup_schema_client.assert_called_with(self.uri)
        self.assertIn(unittest.mock.call().execute_backup_function("testdate"), mock_backup_schema_client.mock_calls)

    @unittest.mock.patch('pychado.tasks.run_select_command')
    def test_run_select(self, mock_run):
        # Checks that database queries are correctly run
        self.assertIs(mock_run, tasks.run_select_command)
        args = ["chado", "extract", "organisms", "-H", "-d", ";", "-o", "testfile", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_run.assert_called_with("organisms", parsed_args, self.uri)

    @unittest.mock.patch('pychado.dbutils.query_and_print')
    @unittest.mock.patch('pychado.queries.set_query_conditions')
    @unittest.mock.patch('pychado.queries.load_query')
    def test_extract_stats(self, mock_load, mock_set, mock_query):
        # Checks that the function providing database statistics is correctly called
        self.assertIs(mock_load, queries.load_query)
        self.assertIs(mock_set, queries.set_query_conditions)
        self.assertIs(mock_query, dbutils.query_and_print)

        args = ["chado", "extract", "stats", "--start_date", "teststartdate", "--end_date", "testenddate",
                "-a", "testorganism", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        mock_load.return_value = "testquery"
        mock_set.return_value = "testquery_with_params"

        tasks.run_select_command(args[2], parsed_args, self.uri)
        mock_load.assert_called_with("stats")
        mock_set.assert_called_with("testquery", organism="testorganism", start_date="teststartdate",
                                    end_date="testenddate")
        mock_query.assert_called_with(self.uri, "testquery_with_params", "", "csv", False, "\t")

        args = ["chado", "extract", "stats", "-H", "-d", ";", "-o", "testfile", "--start_date", "teststartdate",
                "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        mock_load.return_value = "testquery"
        mock_set.return_value = "testquery_with_params"

        tasks.run_select_command(args[2], parsed_args, self.uri)
        mock_load.assert_called_with("stats")
        mock_set.assert_called_with("testquery", organism=None, start_date="teststartdate",
                                    end_date=utils.current_date())
        mock_query.assert_called_with(self.uri, "testquery_with_params", "testfile", "csv", True, ";")

    @unittest.mock.patch('pychado.dbutils.query_and_print')
    @unittest.mock.patch('pychado.io.direct.DirectIOClient')
    def test_extract_organisms(self, mock_client, mock_query):
        # Checks that the function extracting organisms is correctly called
        self.assertIs(mock_client, direct.DirectIOClient)
        self.assertIs(mock_query, dbutils.query_and_print)

        args = ["chado", "extract", "organisms", "--public_only", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        mock_client_obj = mock_client.return_value
        mock_client_obj.configure_mock(**{"select_organisms.return_value": "testquery"})

        tasks.run_select_command(args[2], parsed_args, self.uri)
        mock_client.assert_called_with(self.uri)
        self.assertIn(unittest.mock.call().select_organisms(True, False), mock_client.mock_calls)
        mock_query.assert_called_with(self.uri, "testquery", "", "csv", False, "\t")

    @unittest.mock.patch('pychado.dbutils.query_and_print')
    @unittest.mock.patch('pychado.queries.set_query_conditions')
    @unittest.mock.patch('pychado.queries.load_query')
    def test_extract_cvterms(self, mock_load, mock_set, mock_query):
        # Checks that the function extracting CV terms is correctly called
        self.assertIs(mock_load, queries.load_query)
        self.assertIs(mock_set, queries.set_query_conditions)
        self.assertIs(mock_query, dbutils.query_and_print)

        args = ["chado", "extract", "cvterms", "--database", "testdatabase", "--vocabulary", "testvocabulary", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        mock_load.return_value = "testquery"
        mock_set.return_value = "testquery_with_params"

        tasks.run_select_command(args[2], parsed_args, self.uri)
        mock_load.assert_called_with("cvterms")
        mock_set.assert_called_with("testquery", vocabulary="testvocabulary", database="testdatabase")
        mock_query.assert_called_with(self.uri, "testquery_with_params", "", "csv", False, "\t")

        tasks.run_select_command("inexistent_specifier", parsed_args, self.uri)
        mock_set.assert_called_with("")

    @unittest.mock.patch('pychado.dbutils.query_and_print')
    @unittest.mock.patch('pychado.queries.set_query_conditions')
    @unittest.mock.patch('pychado.queries.load_query')
    def test_extract_genedb_products(self, mock_load, mock_set, mock_query):
        # Checks that the function extracting GeneDB products is correctly called
        self.assertIs(mock_load, queries.load_query)
        self.assertIs(mock_set, queries.set_query_conditions)
        self.assertIs(mock_query, dbutils.query_and_print)

        args = ["chado", "extract", "genedb_products", "-a", "testorganism", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        mock_load.return_value = "testquery"
        mock_set.return_value = "testquery_with_params"
        mock_query.return_value = ['a', 'b', 'c']

        tasks.run_select_command(args[2], parsed_args, self.uri)
        mock_load.assert_called_with("genedb_products")
        mock_set.assert_called_with("testquery", organism="testorganism")
        mock_query.assert_called_with(self.uri, "testquery_with_params", "", "csv", False, "\t")

    @unittest.mock.patch('pychado.tasks.run_insert_command')
    def test_run_insert(self, mock_run):
        # Checks that database inserts are correctly run
        self.assertIs(mock_run, tasks.run_insert_command)
        args = ["chado", "insert", "organism", "-g", "testgenus", "-s", "testspecies", "-a", "testorganism", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_run.assert_called_with("organism", parsed_args, self.uri)

    @unittest.mock.patch('pychado.tasks.run_delete_command')
    def test_run_delete(self, mock_run):
        # Checks that database inserts are correctly run
        self.assertIs(mock_run, tasks.run_delete_command)
        args = ["chado", "delete", "organism", "-a", "testorganism", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_run.assert_called_with("organism", parsed_args, self.uri)

    @unittest.mock.patch('pychado.io.direct.DirectIOClient')
    def test_insert_organism(self, mock_client):
        # Checks that the function inserting organisms is correctly called
        self.assertIs(mock_client, direct.DirectIOClient)
        args = ["chado", "insert", "organism", "-g", "testgenus", "-s", "testspecies", "-a", "testorganism", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)

        tasks.run_insert_command(args[2], parsed_args, self.uri)
        mock_client.assert_called_with(self.uri)
        self.assertIn(unittest.mock.call().insert_organism("testgenus", "testspecies", "testorganism",
                                                           None, None, None, None), mock_client.mock_calls)

        mock_client.reset_mock()
        tasks.run_insert_command("inexistent_specifier", parsed_args, self.uri)
        self.assertEqual([unittest.mock.call(self.uri)], mock_client.mock_calls)

    @unittest.mock.patch('pychado.io.direct.DirectIOClient')
    def test_delete_organism(self, mock_client):
        # Checks that the function inserting organisms is correctly called
        self.assertIs(mock_client, direct.DirectIOClient)
        args = ["chado", "delete", "organism", "-a", "testorganism", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)

        tasks.run_delete_command(args[2], parsed_args, self.uri)
        mock_client.assert_called_with(self.uri)
        self.assertIn(unittest.mock.call().delete_organism("testorganism"), mock_client.mock_calls)

        mock_client.reset_mock()
        tasks.run_delete_command("inexistent_specifier", parsed_args, self.uri)
        self.assertEqual([unittest.mock.call(self.uri)], mock_client.mock_calls)

    @unittest.mock.patch('pychado.tasks.run_import_command')
    def test_run_import(self, mock_run):
        # Checks that database imports are correctly run
        self.assertIs(mock_run, tasks.run_import_command)
        args = ["chado", "import", "essentials", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_run.assert_called_with("essentials", parsed_args, self.uri)

    @unittest.mock.patch('pychado.io.essentials.EssentialsClient')
    def test_import_essentials(self, mock_client):
        # Checks that the function importing essentials into the database is correctly called
        self.assertIs(mock_client, essentials.EssentialsClient)
        args = ["chado", "import", "essentials", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_import_command(args[2], parsed_args, self.uri)
        mock_client.assert_called_with(self.uri, False)
        self.assertIn(unittest.mock.call().load(), mock_client.mock_calls)

        mock_client.reset_mock()
        tasks.run_import_command("inexistent_specifier", parsed_args, self.uri)
        mock_client.assert_not_called()

    @unittest.mock.patch('pychado.utils.download_file')
    @unittest.mock.patch('pychado.io.ontology.OntologyClient')
    def test_import_ontology(self, mock_client, mock_download):
        # Checks that the function importing an ontology into the database is correctly called
        self.assertIs(mock_client, ontology.OntologyClient)
        self.assertIs(mock_download, utils.download_file)
        args = ["chado", "import", "ontology", "-f", "testfile", "-A", "testauthority", "-F", "owl", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_import_command(args[2], parsed_args, self.uri)
        mock_download.assert_not_called()
        mock_client.assert_called_with(self.uri, False)
        self.assertIn(unittest.mock.call().load("testfile", "owl", "testauthority"), mock_client.mock_calls)

        mock_client.reset_mock()
        mock_download.return_value = "downloaded_file"
        args = ["chado", "import", "ontology", "-V", "-u", "testurl", "-A", "testauthority", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_import_command(args[2], parsed_args, self.uri)
        mock_download.assert_called_with("testurl")
        mock_client.assert_called_with(self.uri, True)
        self.assertIn(unittest.mock.call().load("downloaded_file", "obo", "testauthority"), mock_client.mock_calls)

    @unittest.mock.patch('pychado.io.gff.GFFImportClient')
    def test_import_gff(self, mock_client):
        # Checks that the function importing a GFF file into the database is correctly called
        self.assertIs(mock_client, gff.GFFImportClient)
        args = ["chado", "import", "gff", "-f", "testfile", "-a", "testorganism", "--fasta", "testfasta",
                "-t", "contig", "--fresh_load", "--force", "--full_genome", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_import_command(args[2], parsed_args, self.uri)
        mock_client.assert_called_with(self.uri, False)
        self.assertIn(unittest.mock.call().load("testfile", "testorganism", "testfasta", "contig", True, True, True),
                      mock_client.mock_calls)

    @unittest.mock.patch('pychado.io.fasta.FastaImportClient')
    def test_import_fasta(self, mock_client):
        # Checks that the function importing a FASTA file into the database is correctly called
        self.assertIs(mock_client, fasta.FastaImportClient)
        args = ["chado", "import", "fasta", "-f", "testfile", "-a", "testorganism", "-t", "contig", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_import_command(args[2], parsed_args, self.uri)
        mock_client.assert_called_with(self.uri, False)
        self.assertIn(unittest.mock.call().load("testfile", "testorganism", "contig"), mock_client.mock_calls)

    @unittest.mock.patch('pychado.io.gaf.GAFImportClient')
    def test_import_gaf(self, mock_client):
        # Checks that the function importing a GAF file into the database is correctly called
        self.assertIs(mock_client, gaf.GAFImportClient)
        args = ["chado", "import", "gaf", "-f", "testfile", "-a", "testorganism", "-L", "protein", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_import_command(args[2], parsed_args, self.uri)
        mock_client.assert_called_with(self.uri, False)
        self.assertIn(unittest.mock.call().load("testfile", "testorganism", "protein"), mock_client.mock_calls)

    @unittest.mock.patch('pychado.tasks.run_export_command')
    def test_run_export(self, mock_run):
        # Checks that database exports are correctly run
        self.assertIs(mock_run, tasks.run_export_command)
        args = ["chado", "export", "fasta", "-f", "testfile", "-a", "testorganism", "-t", "proteins", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_command_with_arguments(args[1], args[2], parsed_args, self.uri)
        mock_run.assert_called_with("fasta", parsed_args, self.uri)

    @unittest.mock.patch('pychado.io.fasta.FastaExportClient')
    def test_export_fasta(self, mock_client):
        # Checks that the function exporting sequences from the database to a FASTA file is correctly called
        self.assertIs(mock_client, fasta.FastaExportClient)
        args = ["chado", "export", "fasta", "-f", "testfile", "-a", "testorganism", "-t", "proteins",
                "-r", "testrelease", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_export_command(args[2], parsed_args, self.uri)
        mock_client.assert_called_with(self.uri, False)
        self.assertIn(unittest.mock.call().export("testfile", "testorganism", "proteins", "testrelease", False, False),
                      mock_client.mock_calls)

    @unittest.mock.patch('pychado.io.gff.GFFExportClient')
    def test_export_gff(self, mock_client):
        # Checks that the function exporting genomic data from the database to a GFF file is correctly called
        self.assertIs(mock_client, gff.GFFExportClient)
        args = ["chado", "export", "gff", "-f", "testfile", "-a", "testorganism", "--export_fasta", "--fasta_file",
                "testfasta", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_export_command(args[2], parsed_args, self.uri)
        mock_client.assert_called_with(self.uri, False)
        self.assertIn(unittest.mock.call().export("testfile", "testorganism", True, "testfasta", False),
                      mock_client.mock_calls)

    @unittest.mock.patch('pychado.io.gaf.GAFExportClient')
    def test_export_gaf(self, mock_client):
        # Checks that the function exporting gene annotation data from the database to a GAF file is correctly called
        self.assertIs(mock_client, gaf.GAFExportClient)
        args = ["chado", "export", "gaf", "-f", "testfile", "-a", "testorganism", "-A", "testauthority",
                "-L", "protein", "testdb"]
        parsed_args = chado_tools.parse_arguments(args)
        tasks.run_export_command(args[2], parsed_args, self.uri)
        mock_client.assert_called_with(self.uri, False)
        self.assertIn(unittest.mock.call().export("testfile", "testorganism", "testauthority", "protein", False),
                      mock_client.mock_calls)


if __name__ == '__main__':
    unittest.main(verbosity=2, buffer=True)



import unittest
from .. import dbutils, utils
from ..io import essentials
from ..orm import base, general, cv, pub


class TestEssentials(unittest.TestCase):
    """Tests various functions used to load an ontology from a file into a database"""

    connection_parameters = utils.parse_yaml(dbutils.default_configuration_file())
    connection_uri = dbutils.random_database_uri(connection_parameters)

    @classmethod
    def setUpClass(cls):
        # Creates a database, establishes a connection, and creates tables
        dbutils.create_database(cls.connection_uri)
        schema_base = base.PublicBase
        schema_metadata = schema_base.metadata
        cls.client = essentials.EssentialsClient(cls.connection_uri)
        schema_metadata.create_all(cls.client.engine, tables=schema_metadata.sorted_tables)

    @classmethod
    def tearDownClass(cls):
        # Drops the database
        dbutils.drop_database(cls.connection_uri, True)

    def tearDown(self):
        # Rolls back all changes to the database
        self.client.session.rollback()

    def test_load_generic_entries(self):
        # Tests the loading of default entries into the main tables of the Chado database
        self.client._load_generic_entries()
        null_db = self.client.query_first(general.Db)                                       # type: general.Db
        self.assertIsNotNone(null_db.db_id)
        self.assertEqual(null_db.name, "null")
        null_dbxref = self.client.query_first(general.DbxRef)                               # type: general.DbxRef
        self.assertIsNotNone(null_dbxref.dbxref_id)
        self.assertEqual(null_dbxref.accession, "null")
        null_cv = self.client.query_first(cv.Cv)                                            # type: cv.Cv
        self.assertIsNotNone(null_cv.cv_id)
        self.assertEqual(null_cv.name, "null")
        null_cvterm = self.client.query_first(cv.CvTerm)                                    # type: cv.CvTerm
        self.assertIsNotNone(null_cvterm.cvterm_id)
        self.assertEqual(null_cvterm.name, "null")
        null_pub = self.client.query_first(pub.Pub)                                         # type: pub.Pub
        self.assertIsNotNone(null_pub.pub_id)
        self.assertEqual(null_pub.uniquename, "null")

    def test_load_relationship_entries(self):
        # Tests the loading of relationship entries into the Chado 'cvterm' table
        self.client._load_relationship_entries()
        relationship_cv = self.client.query_first(cv.Cv, name="relationship")               # type: cv.Cv
        self.assertIsNotNone(relationship_cv.cv_id)
        is_a_term = self.client.query_first(cv.CvTerm, name="is_a")                         # type: cv.CvTerm
        self.assertIsNotNone(is_a_term.cvterm_id)
        self.assertEqual(is_a_term.cv_id, relationship_cv.cv_id)
        self.assertEqual(is_a_term.is_relationshiptype, 1)
        self.client._load_relationship_entries("has_part")
        has_part_term = self.client.query_first(cv.CvTerm, name="has_part")                 # type: cv.CvTerm
        self.assertIsNotNone(has_part_term.cvterm_id)

    def test_load_synonymtype_entries(self):
        # Tests the loading of synonym type entries into the Chado 'cvterm' table
        self.client._load_synonymtype_entries()
        synonym_type_cv = self.client.query_first(cv.Cv, name="synonym_type")               # type: cv.Cv
        self.assertIsNotNone(synonym_type_cv.cv_id)
        narrow_cvterm = self.client.query_first(cv.CvTerm, name="narrow")                   # type: cv.CvTerm
        self.assertIsNotNone(narrow_cvterm.cvterm_id)
        self.assertEqual(narrow_cvterm.cv_id, synonym_type_cv.cv_id)

    def test_load_cvterm_property_type_entries(self):
        # Tests the loading of cvterm property type entries into the Chado 'cvterm' table
        self.client._load_cvterm_property_type_entries()
        property_type_cv = self.client.query_first(cv.Cv, name="cvterm_property_type")      # type: cv.Cv
        self.assertIsNotNone(property_type_cv.cv_id)
        symmetric_cvterm = self.client.query_first(cv.CvTerm, name="is_symmetric")          # type: cv.CvTerm
        self.assertIsNotNone(symmetric_cvterm.cvterm_id)
        self.assertEqual(symmetric_cvterm.cv_id, property_type_cv.cv_id)
        self.assertEqual(symmetric_cvterm.is_relationshiptype, 0)
        disjoint_cvterm = self.client.query_first(cv.CvTerm, name="disjoint_from")          # type: cv.CvTerm
        self.assertIsNotNone(disjoint_cvterm.cvterm_id)
        self.assertEqual(disjoint_cvterm.cv_id, property_type_cv.cv_id)
        self.assertEqual(disjoint_cvterm.is_relationshiptype, 1)

    def test_load_feature_property_entries(self):
        # Tests the loading of feature property entries into the Chado 'cvterm' table
        self.client._load_feature_property_entries()
        feature_property_cv = self.client.query_first(cv.Cv, name="feature_property")       # type: cv.Cv
        self.assertIsNotNone(feature_property_cv.cv_id)
        comment_cvterm = self.client.query_first(cv.CvTerm, name="comment")                 # type: cv.CvTerm
        self.assertIsNotNone(comment_cvterm.cvterm_id)
        self.assertEqual(comment_cvterm.cv_id, feature_property_cv.cv_id)

    def test_load_genedb_synonymtype_entries(self):
        # Tests the loading of specific GeneDB synonym type entries into the Chado 'cvterm' table
        self.client._load_genedb_synonymtype_entries()
        synonym_type_cv = self.client.query_first(cv.Cv, name="genedb_synonym_type")        # type: cv.Cv
        self.assertIsNotNone(synonym_type_cv.cv_id)
        systematic_id_cvterm = self.client.query_first(cv.CvTerm, name="systematic_id")     # type: cv.CvTerm
        self.assertIsNotNone(systematic_id_cvterm.cvterm_id)
        self.assertEqual(systematic_id_cvterm.cv_id, synonym_type_cv.cv_id)

    def test_load_genedb_misc_entries(self):
        # Tests the loading of GeneDB-specific entries into the Chado 'cvterm' table
        self.client._load_genedb_misc_entries()
        misc_dv = self.client.query_first(cv.Cv, name="genedb_misc")                        # type: cv.Cv
        self.assertIsNotNone(misc_dv.cv_id)
        top_level_seq_cvterm = self.client.query_first(cv.CvTerm, name="top_level_seq")     # type: cv.CvTerm
        self.assertIsNotNone(top_level_seq_cvterm.cvterm_id)
        self.assertEqual(top_level_seq_cvterm.cv_id, misc_dv.cv_id)



import unittest.mock
import pkg_resources
from .. import queries


class TestQueries(unittest.TestCase):
    """Tests the loading and modification of stored SQL queries and templates"""

    def test_sql_resources(self):
        # Checks that all required resources are available
        self.assertTrue(pkg_resources.resource_isdir("pychado", "sql"))
        self.assertTrue(pkg_resources.resource_exists("pychado", "sql/extract_stats.sql"))
        self.assertTrue(pkg_resources.resource_exists("pychado", "sql/extract_organisms.sql"))
        self.assertTrue(pkg_resources.resource_exists("pychado", "sql/extract_cvterms.sql"))
        self.assertTrue(pkg_resources.resource_exists("pychado", "sql/extract_genedb_products.sql"))

    def test_load_query(self):
        # Checks that the templates for the 'chado extracts' queries are correctly loaded
        query = queries.load_query("organisms")
        self.assertIn("SELECT", query)

        query = queries.load_query("cvterms")
        self.assertIn("SELECT", query)

        query = queries.load_query("genedb_products")
        self.assertIn("SELECT", query)

        query = queries.load_query("stats")
        self.assertIn("SELECT", query)

        query = queries.load_query("non_existent_specifier")
        self.assertEqual(query, "")

    def test_bind_parameters(self):
        # Checks if a parameter is correctly bound to a query
        query = "SELECT * FROM testtable WHERE mycolumn = :myvalue"
        query_with_params = queries.bind_parameters(query, myvalue="testvalue")
        query_with_params_inlined = str(query_with_params.compile(compile_kwargs={"literal_binds": True}))
        self.assertEqual(query_with_params_inlined, "SELECT * FROM testtable WHERE mycolumn = 'testvalue'")

        query = "SELECT * FROM testtable WHERE col1 = :val1 AND col2 = :val2"
        query_with_params = queries.bind_parameters(query, val1=12, val2=True)
        query_with_params_inlined = str(query_with_params.compile(compile_kwargs={"literal_binds": True}))
        self.assertEqual(query_with_params_inlined, "SELECT * FROM testtable WHERE col1 = 12 AND col2 = true")

    @unittest.mock.patch("pychado.queries.bind_parameters")
    @unittest.mock.patch("pychado.queries.set_vocabulary_condition")
    @unittest.mock.patch("pychado.queries.set_database_condition")
    @unittest.mock.patch("pychado.queries.set_organism_condition")
    def test_set_conditions(self, mock_organism, mock_database, mock_vocabulary, mock_bind):
        # Checks if placeholders in a query are replaced with actual conditions
        self.assertIs(mock_organism, queries.set_organism_condition)
        self.assertIs(mock_database, queries.set_database_condition)
        self.assertIs(mock_vocabulary, queries.set_vocabulary_condition)
        self.assertIs(mock_bind, queries.bind_parameters)

        query = "SELECT * FROM testtable"
        queries.set_query_conditions(query)
        mock_organism.assert_not_called()
        mock_database.assert_not_called()
        mock_vocabulary.assert_not_called()
        mock_bind.assert_called_with(query)

        mock_bind.reset_mock()
        mock_organism.return_value = "modified_query"
        mock_database.return_value = "further_modified_query"
        mock_vocabulary.return_value = "final_query"
        queries.set_query_conditions(query, organism="testorganism", database="testdatabase",
                                     vocabulary="testvocabulary")
        mock_organism.assert_called_with(query, "testorganism")
        mock_database.assert_called_with("modified_query", "testdatabase")
        mock_vocabulary.assert_called_with("further_modified_query", "testvocabulary")
        mock_bind.assert_called_with("final_query", organism="testorganism", database="testdatabase",
                                     vocabulary="testvocabulary")

    def test_set_organism_condition(self):
        # Checks that a query placeholder is correctly replaced
        query = "SELECT * FROM testtable WHERE :ORGANISM_CONDITION"
        self.assertEqual(queries.set_organism_condition(query, ""),
                         "SELECT * FROM testtable WHERE TRUE")
        self.assertEqual(queries.set_organism_condition(query, "testorganism"),
                         "SELECT * FROM testtable WHERE abbreviation = :organism")

    def test_set_vocabulary_condition(self):
        # Checks that a query placeholder is correctly replaced
        query = "SELECT * FROM testtable WHERE :CV_CONDITION"
        self.assertEqual(queries.set_vocabulary_condition(query, ""),
                         "SELECT * FROM testtable WHERE TRUE")
        self.assertEqual(queries.set_vocabulary_condition(query, "testvocabulary"),
                         "SELECT * FROM testtable WHERE cv.name = :vocabulary")

    def test_set_database_condition(self):
        # Checks that a query placeholder is correctly replaced
        query = "SELECT * FROM testtable WHERE :DB_CONDITION"
        self.assertEqual(queries.set_database_condition(query, ""),
                         "SELECT * FROM testtable WHERE TRUE")
        self.assertEqual(queries.set_database_condition(query, "testdatabase"),
                         "SELECT * FROM testtable WHERE db.name = :database")


if __name__ == '__main__':
    unittest.main(verbosity=2, buffer=True)
